/*
  A smoke test for cphstl::vector; this test is taken (with some minor
  modifications) from the book [P. J. Plauger, Alexander A. Stepanov,
  Meng Lee, and David R.  Musser, The C++ Standard Template Library,
  Prentice Hall PTR (2001)].

  Author: Jyrki Katajainen, April 2008
  Modified by Bo Simonsen, November 2008
*/

#include <memory> // defines std::allocator
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <cassert> // defines assert
#include <vector> // defines std::vector
#include <iostream>

#ifdef IV
#include "iv_dynamic_array.h++"
#else
  #ifndef STD
  #include "dynamic_array.h++"
  #endif
#endif

#include "entry_iterator.h++"
#include "vector.h++" // defines cphstl::vector

template <typename T>
inline void ignore_unused_variable_warning(T const&) {
}

template <typename V>
void test_vector(V* a, std::size_t n) {

  assert(n > 0);
  //  assert(cphstl::is_sorted(a, a + n);
  typedef std::allocator<V> A;  

#ifdef STD 
#define KERNEL std::vector<V, A> 
#define ITERATOR 
#else
  #ifdef IV
  #define KERNEL cphstl::iv_dynamic_array<V, A>
  #define ITERATOR ,cphstl::entry_iterator< KERNEL, cphstl::iv_dynamic_array_entry<V,A>, false>, cphstl::entry_iterator< KERNEL, cphstl::iv_dynamic_array_entry<V,A>, true>
  #else 
  #define KERNEL cphstl::dynamic_array<V, A>
  #define ITERATOR ,cphstl::entry_iterator< KERNEL, cphstl::dynamic_array_entry<V,A>, false>, cphstl::entry_iterator< KERNEL, cphstl::dynamic_array_entry<V,A>, true>
  #endif
#endif

  typedef cphstl::vector<V, A, KERNEL ITERATOR > container;

  V const x = a[0];
  V y = x;
  V const z = a[n - 1];
  typename container::allocator_type* p_alloc = (A*)0;
  ignore_unused_variable_warning(p_alloc);
  typename container::pointer p_ptr = (V*)0;
  ignore_unused_variable_warning(p_ptr);
  typename container::const_pointer p_cptr = (V const*)0;
  ignore_unused_variable_warning(p_cptr);
  typename container::reference p_ref = y;
  ignore_unused_variable_warning(p_ref);
  typename container::const_reference p_cref = (V const&) y;
  ignore_unused_variable_warning(p_cref);
  typename container::size_type* p_size = (std::size_t*) 0;
  ignore_unused_variable_warning(p_size);
  typename container::difference_type* p_diff = (std::ptrdiff_t*) 0;
  ignore_unused_variable_warning(p_diff);
  typename container::value_type* p_value = (V*) 0;
  ignore_unused_variable_warning(p_value);

  container v0;
  assert(v0.empty());
  assert(v0.size() == 0);
  A al = v0.get_allocator();
  container v0a(al);
  assert(v0a.size() == 0);
  assert(v0a.get_allocator() == al);
  container v1(5);
  assert(v1.size() == 5);  
  assert(v1.back() == V());
  container v1a(6, x);
  assert(v1a.size() == 6);  
  assert(v1a.back() == x);
  container v1b(7, x, al);
  assert(v1b.size() == 7);  
  assert(v1b.back() == x);
  container v2(v1a);
  assert(v2.size() == 6);  
  assert(v2.front() == x);
  container v3(v1a.begin(), v1a.end());
  assert(v3.size() == 6);
  assert(v3.front() == x);
  const container v4(v1a.begin(), v1a.end(), al);
  assert(v4.size() == 6);  
  assert(v4.front() == x);
  v0 = v4;
  assert(v0.size() == 6);  
  assert(v0.front() == x);
  assert(v0[0] == x);
  assert(v0.at(5) == x);

  v0.reserve(12);
  assert(12 <= v0.capacity());
  v0.resize(8);
  assert(v0.size() == 8);  
  assert(v0.back() == V());
  v0.resize(10, z);
  assert(v0.size() == 10);  
  assert(v0.back() == z);
  assert(v0.size() <= v0.max_size());  

  typedef typename container::iterator iterator;
  iterator p_it(v0.begin());
  typedef typename container::const_iterator const_iterator;
  const_iterator p_cit(v4.begin());
  typedef typename container::reverse_iterator reverse_iterator;
  reverse_iterator p_rit(v0.rbegin());
  typedef typename container::const_reverse_iterator const_reverse_iterator;
  const_reverse_iterator p_crit(v4.rbegin());
  assert(*p_it == x);
  p_it = v0.end();
  --p_it;
  assert(*p_it == z);
  assert(*p_cit == x);
  p_cit = v4.end();
  --p_cit;
  assert(*p_cit == x);
  assert(*p_rit == z);
  p_rit = v0.rend();
  --p_rit;
  assert(*p_rit == x);  
  assert(*p_crit == x);
  p_crit = v4.rend();
  --p_crit;
  assert(*p_crit == x);  
  
  assert(v0.front() == x);
  assert(v4.front() == x);
  v0.push_back(a[0]);
  assert(v0.back() == a[0]);
  v0.pop_back();
  assert(v0.back() == z);  
  assert(v4.back() == x);  
//v0.push_front(z);
//assert(v0.front() == z);  
//v0.pop_front();
//assert(v0.front() == x);  

  v0.assign(v4.begin(), v4.end());
  assert(v0.size() == v4.size());
  assert(v0.front() == v4.front());
  v0.assign(4, z);
  assert(v0.size() == 4);
  assert(v0.front() == z);  
  p_cit = v0.insert(v0.begin(), x);
  assert(*p_cit == x);
  ++p_cit;
  assert(v0.front() == x);
  assert(*p_cit == z);
  v0.insert(v0.begin(), 1, x);
  p_cit = v0.begin();
  assert(*p_cit == x);
  p_cit++;
  assert(*p_cit == x);
  ++p_cit;
  assert(*p_cit == z);
  v0.insert(v0.end(), v4.begin(), v4.end());
  assert(v0.back() == v4.back());
  v0.insert(v0.end(), a, a + n);
  assert(v0.back() == z);
  v0.erase(v0.begin());  
  p_cit = v0.begin();
  assert(*p_cit == x);
  p_cit++;
  assert(*p_cit == z);
  p_it = ++v0.begin();
  v0.erase(v0.begin(), p_it);
  assert(v0.front() == z);

  v0.clear();
  assert(v0.empty());
  v0.swap(v1);
  assert(!v0.empty());
  assert(v1.empty());
  swap(v0, v1);
  assert(v0.empty());
  assert(!v1.empty());
  assert(v1 == v1);
  assert(v0 < v1);
  assert(v0 != v1);
  assert(v1 > v0);
  assert(v0 <= v1);
  assert(v1 >= v0);

}

int main(int, char**) {
  int a[] = {1, 2, 3};
  test_vector<int>(a, sizeof(a) / sizeof(a[0]));

  char b[] ="abc";
  test_vector<char>(b, sizeof(b) / sizeof(b[0]) - 1); // \0 removed
  return 0;
}
