/*

  Implementation of the AVL tree

  Authors: Bo Simonsen, May 2008, December 2008

*/

/*
 * Copyright (c) 2003
 * Stephan Lynge, Department of Computer Science University of Copenhagen
 *
 * Copyright (c) 2002, 2003
 * Herve Bronnimann, Polytechnic University
 *
 * Copyright (c) 1996,1997
 * Silicon Graphics Computer Systems, Inc.
 *
 * Copyright (c) 1994
 * Hewlett-Packard Company
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation.  Neither the author nor Polytechnic
 * University, nor Silicon Graphics, nor Hewlett Packard, makes any
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 */

#include <iostream>

namespace cphstl {
 
  /* (C, A) constructor */
 
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  avl_tree<K, V, F, C, A, N, is_multiset>::avl_tree(const key_compare& k, const allocator_type& a) {

    (*this).comparator = k;
    (*this).allocator = a;

    (*this).header = (*this).make_node(); 
    (*this).init(); 
  }

  /* Destructor */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  avl_tree<K, V, F, C, A, N, is_multiset>::~avl_tree() {
    (*this).destroy_tree((*this).root()); 
    (*this).free_node((*this).header); 
  }

  /* get_allocator */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::allocator_type
  avl_tree<K, V, F, C, A, N, is_multiset>::get_allocator() const {
    return allocator_type();
  }

  /* key_comp */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::key_compare
  avl_tree<K, V, F, C, A, N, is_multiset>::key_comp() const {
    return key_compare();
  }

  /* value_comp */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::value_compare
  avl_tree<K, V, F, C, A, N, is_multiset>::value_comp() const {
    return value_compare();
  }

  /* Init */
  /* EXCEPTION SAFE */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void avl_tree<K, V, F, C, A, N, is_multiset>::init()
  {
    (*this).root() = 0;
    (*this).leftmost() = (*this).header;
    (*this).rightmost() = (*this).header;
    (*this).node_count = 0;
  }

  /* make_node */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::make_node() {
    return (*this).allocator.allocate(1);
  }

  /* free_node */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::free_node(node_type* p) {
    return (*this).allocator.deallocate(p, 1);
  }

  /* construct - implementing "placement new" */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::construct(pointer p, const value_type& v) const {
    new((void*)p) value_type(v); 
  }

  /* destroy - implementing "placement delete" */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::destroy(pointer p) const {
    (*p).~value_type();
  }
   
  /* root */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*&
  avl_tree<K, V, F, C, A, N, is_multiset>::root() const {
    return (*(*this).header).parent();
  }

  /* leftmost */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*&
  avl_tree<K, V, F, C, A, N, is_multiset>::leftmost() const {
    return (*(*this).header).left();
  }

  /* rightmost */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*&
  avl_tree<K, V, F, C, A, N, is_multiset>::rightmost() const {
    return (*(*this).header).right();
  }

  /* tree_maximum */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::tree_maximum(node_type* x) const {
    while ((*x).right() != 0) {
      x = (*x).right();
    }
    return x;
  }

  /* tree_minimum */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::tree_minimum(node_type* x) const {
    while ((*x).left() != 0) {
      x = (*x).left();
    }
    return x;
  }


  /* create_node - Allocates and construct a node with value_type v */
  /* EXCEPTION SAFE */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::create_node(const value_type& v) {

    node_type* tmp;

    try {
      tmp = (*this).make_node();
      (*this).construct(&(*tmp).content(), v);
    }
    catch (...) {
      if(tmp != 0) {
        (*this).free_node(tmp);
      }
      throw;
    }

    (*tmp).left() = 0;
    (*tmp).right() = 0;
    (*tmp).parent() = 0;
    (*tmp).predecessor() = 0;
    (*tmp).successor() = 0;

    return tmp;
     
  }

  /* destroy_node - Destruct and destroy the node */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void avl_tree<K, V, F, C, A, N, is_multiset>::destroy_node(node_type* p) {
    (*this).destroy(&(*p).content());
    (*this).free_node(p);
  }

  /* destroy_tree - Will destruct the entire tree */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void avl_tree<K, V, F, C, A, N, is_multiset>::destroy_tree(node_type* x) {
    while (x != 0) {
      (*this).destroy_tree((*x).right());
      node_type* y = (*x).left();
      (*this).destroy_node(x);
      x = y;
    }
      
  }

  /* build_tree
     This O(n) routine builds a perfect balanced tree with h = O(lg n) given a vector (array) of
     of pointers to nodes. The routine is based on the devide-and-conquer strategy.
     EXCEPTION SAFE.
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::build_tree(node_type* nodes[], int nodecount) {
      
    if(nodecount == 0)
      return 0;

    unsigned int median = (nodecount) / 2;

    /* Build the left subtree */
    (*nodes[median]).left() = build_tree(nodes, median);
    if((*nodes[median]).left() != 0)  {
      (*(*nodes[median]).left()).parent() = nodes[median];
    }
 
    /* Build the right subtree */
    (*nodes[median]).right() = build_tree(nodes+median+1, nodecount - median - 1);
    if((*nodes[median]).right() != 0) {
      (*(*nodes[median]).right()).parent() = nodes[median];
    }

    /* Adjust the balance O(1) */
    if((*nodes[median]).left() != 0 && (*nodes[median]).right() != 0) {
      (*nodes[median]).balance() = 0;
    }
    else if((*nodes[median]).left() != 0) {
      (*nodes[median]).balance() = -1;
    }
    else if((*nodes[median]).right() != 0) {
      (*nodes[median]).balance() = 1;
    }
    else {
      (*nodes[median]).balance() = 0;
    }

    return nodes[median];
  }

  /* copy_nodes

     The actual copying in O(n) time, to be precise it's done in 3n time. It doesn't edit class 
     variables simply takes a pointer to start and end position and numbers of elements, and 
     returning a pointer to the new tree
     EXCEPTION SAFE.
   */
 
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::copy_nodes(node_type* begin, node_type* end, int nodecount) {
      
    node_type* nodes[nodecount];
    int i = 0;

    try {
      for(node_type* p = begin; p != end; p = (*p).successor()) {
        nodes[i] = (*this).create_node((*p).content());
        ++i;  
      }
    } catch(...) {
      for (int j=0; j!=i; ++j) {
        (*this).destroy_node(nodes[j]);
      }
      throw;
    }
    /* Adjust the associated data structure */
    (*nodes[0]).predecessor() = 0;
    (*nodes[0]).successor() = nodes[1];
    
    for(i = 1; i < nodecount - 1; ++i) {
      (*nodes[i]).predecessor() = nodes[i-1];
      (*nodes[i]).successor() = nodes[i+1];
    }

    (*nodes[nodecount-1]).predecessor() = nodes[nodecount - 2];
    (*nodes[nodecount-1]).successor() = 0;
        
    return (*this).build_tree(nodes, nodecount);
  }

  /* copy
  
     Just a wrapper function for copying 
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void avl_tree<K, V, F, C, A, N, is_multiset>::copy(const avl_tree& x) {
    (*this).root() = (*this).copy_nodes(x.begin(), x.end(), x.node_count);
    (*(*this).root()).parent() = (*this).header;
    (*this).leftmost() = (*this).tree_minimum(root());
    (*(*this).leftmost()).predecessor() = (*this).header;
    (*this).rightmost() = (*this).tree_maximum(root());
    (*(*this).rightmost()).successor() = (*this).header;
    (*(*this).header).predecessor() = (*this).rightmost();
    (*(*this).header).successor() = (*this).leftmost();
    (*this).node_count = x.node_count;
  }
    
  /* Copy constructor
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  avl_tree<K, V, F, C, A, N, is_multiset>::avl_tree(const avl_tree& x) {

    try {
      (*this).header = (*this).make_node();
     	(*this).init();
      if(x.root() != 0) {
        (*this).copy(x);
      }
    }
    catch (...) {
      if ((*this).header != 0) {
        (*this).destroy_tree((*this).root());
        (*this).destroy_node((*this).header);
      }
      throw;
    }

    /* Bug: should be pointers, copy construction can throw */
    (*this).allocator = x.allocator;
    (*this).comparator = x.comparator;
  }

  /* assignment operator
     EXCEPTION SAFE - but needs verification 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  avl_tree<K, V, F, C, A, N, is_multiset>&
  avl_tree<K, V, F, C, A, N, is_multiset>::operator=(const avl_tree& x) {
    if (this != &x) {
      node_type* old_header = (*this).header;
      try {
        (*this).header = (*this).make_node();
      }
      catch(...) {
        (*this).header = old_header;
        throw;
      }

      try {
        (*this).init();

        if(x.root() != 0) {
          (*this).copy(x);
        }
      }
      catch(...) {
        (*this).destroy_tree((*this).root());
        (*this).destroy_node((*this).header);
        (*this).header = old_header;
        throw;
      }

      /* It's now safe to free the old tree */
      (*this).destroy_tree((*old_header).parent());
      (*this).destroy_node(old_header);

      (*this).allocator = x.allocator;
      (*this).comparator = x.comparator;

    }
    return *this;

  }

  /* size */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::size_type
  avl_tree<K, V, F, C, A, N, is_multiset>::size() const {
    return (*this).node_count;
  }

  /* max_size */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::size_type
  avl_tree<K, V, F, C, A, N, is_multiset>::max_size() const {
    return (*this).allocator.max_size();
  }

  /* begin */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::begin() const {
    return (*this).leftmost();
  }

  /* end */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::end() const {
    return (*this).header;
  }

  /* insert(v) */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename cphstl::if_then_else<is_multiset, 
                                 typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*, 
                                 std::pair<typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*, bool> 
                                >::type
  avl_tree<K, V, F, C, A, N, is_multiset>::insert(const value_type& v) {
    return (*this).insert_dispatch(v, cphstl::bool2type<is_multiset>());
  }

  /* insert(pos, v) */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::insert(node_type* position, const value_type& v) {
    return (*this).insert_dispatch(position, v, cphstl::bool2type<is_multiset>());
  }

  /* insert(I, I) */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  template <typename I>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::insert(I x, I y) {

    if(is_sorted(x, y, (*this).value_comp())) {
      (*this).insert_sorted(x,y);
      return;
    }

    (*this).insert_dispatch(x, y, cphstl::bool2type<is_multiset>());
  }

  /* _insert - Where the actual insertion is done, 
     after that rebalancing is done 
 
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::_insert(node_type* x, node_type* y, const value_type& v, node_type* pred, node_type* succ) {

    functor_type f;
    bool comp;

    node_type* z = (*this).create_node(v);

    node_type* pred_next = (*pred).successor();
    node_type* succ_prev = (*succ).predecessor();

    (*z).predecessor() = pred;
    (*z).successor() = succ;

    try {
      /* This is nessersary since we don't construct the data field for
         the header, the same for successor. (Valgrind found this problem) */

      key_type pred_key = f((*pred).content());
      key_type succ_key = f((*succ).content());

      key_type z_key = f((*z).content());

      if(pred == (*this).header) {
         (*pred).successor() = z;
      }
      /* pred_key < z_key || pred_key == z_key */
      else if((*this).comparator(pred_key, z_key) || 
              (!(*this).comparator(pred_key, z_key) && !(*this).comparator(z_key, pred_key))) {
         (*pred).successor() = z; 
      }
      if(succ == (*this).header) {
        (*succ).predecessor() = z;
      }
      /* succ_key > z_key || succ_key == z_key */
      else if(!(*this).comparator(succ_key, z_key) || 
              (!(*this).comparator(succ_key, z_key) && !(*this).comparator(z_key, succ_key))) {
        (*succ).predecessor() = z;
      }
      key_type v_key = f(v);
      key_type y_key = f((*y).content());
      comp = (*this).comparator(v_key, y_key);

    }
    catch(...) {
      (*this).destroy_node(z);
      (*pred).successor() = pred_next;
      (*succ).predecessor() = succ_prev;
      throw;
    }


    if (y == (*this).header || x != 0 || comp) {
      (*y).left() = z;               // also makes m_leftmost() = z 
                                         // when y == m_header
      if (y == (*this).header) {
        (*this).root() = z;
        (*this).rightmost() = z;
      }
      else if (y == (*this).leftmost()) {
        (*this).leftmost() = z;   // maintain m_leftmost() pointing to min node
      }
    }
    else {
      (*y).right() = z;
      if (y == (*this).rightmost()) {
        (*this).rightmost() = z;  // maintain m_rightmost() pointing to max node
      }
    }

    (*z).parent() = y;
    (*z).left() = 0;
    (*z).right() = 0;
    (*z).balance() = 0; // the new node has no children

    // After insertion of a node, we have to rebalance (unless it is the first node)
    if( (*this).root() != z ) {
      short link_side = (*this).get_side( z, (*(*this).header).parent());
      (*this).rebalance((*z).parent(), (*(*this).header).parent(), 1, link_side);
    }

    ++(*this).node_count;
    return z;
  }

  /* rebalance - Done buttom-up.
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::rebalance(node_type* x, node_type*& root, int cmd, short side) {
    while( side != 0 ) {
      if( (*x).balance() != cmd * side ) {
        (*x).balance() = (*x).balance() + cmd * side;
      }
      else {
        side = (*x).balance();
        node_type* y = (*x).left();
        if( (*x).balance() == 1 ) {
          y = (*x).right();
        }
        if( (*y).balance() == -side ) {
       	  node_type* z = (*y).right();
       	  if( side == 1 ) {
       	    z = (*y).left();
          }
       	  if( (*z).balance() == -side ) {
       	    (*x).balance() = 0;
       	    (*y).balance() = side;
       	    (*z).balance() = 0;
       	  }
       	  else {
            if( (*z).balance() == side ) {
              (*x).balance() = -side;
              (*y).balance() = 0;
              (*z).balance() = 0;
       	    }
       	    else {
              (*x).balance() = 0;
             	(*y).balance() = 0;
       	    }
       	  }
       	  if( side == -1 ) {
       	    (*this).rotate_left(y, root);
          }
       	  else {
       	    (*this).rotate_right(y, root);
          }
       	}
        else {
       	  if( (*y).balance() == side ) {
       	    (*x).balance() = 0;
       	    (*y).balance() = 0;
       	  }
       	  else {
       	    if( (*y).balance() == 0 ) {
              (*x).balance() = side;
              (*y).balance() = -side;
       	    }
       	  }
       	}
       	if( side == -1 ) {
       	  y = (*x).left();
       	  (*this).rotate_right(x, root);
       	}
       	else {
          y = (*x).right();
       	  (*this).rotate_left(x, root);
        }
       	x = y;
      }
      if( ( cmd == 1 && (*x).balance() == 0 ) || ( cmd == -1 && (*x).balance() != 0 ) ) {
        break;
      }
	
      side = (*this).get_side( x, root);	
      x = (*x).parent();
    }
  }

  /* rotate_left - Performs a left rotation of 2 nodes and their subtrees.
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::rotate_left(node_type* x, node_type*& root)
  {
    node_type* y = (*x).right();
    (*x).right() = (*y).left();

    if ((*y).left() !=0) {
      (*(*y).left()).parent() = x;
    }

    (*y).parent() = (*x).parent();
    if (x == root) {
      root = y;
    }

    else if (x == (*(*x).parent()).left()) {
      (*(*x).parent()).left() = y;
    }
    else {
      (*(*x).parent()).right() = y;
    }

    (*y).left() = x;
    (*x).parent() = y;
      
  }
 
  /* rotate_right - Performs a right rotation of 2 nodes and their subtrees (opposite of rotate_left)
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::rotate_right(node_type* x, node_type*& root)
  {
    node_type* y = (*x).left();
    (*x).left() = (*y).right();

    if ((*y).right() !=0) {
      (*(*y).right()).parent() = x;
    }

    (*y).parent() = (*x).parent();

    if (x == root) {
      root = y;
    }
    else if (x == (*(*x).parent()).right()) {
      (*(*x).parent()).right() = y;
    }
    else {
      (*(*x).parent()).left() = y;
    }

    (*y).right() = x;
    (*x).parent() = y;
 
  } 

  /* 	get_side - is the node attached to the left or right side of the parent
      EXCEPTION SAFE
   */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  short
  avl_tree<K, V, F, C, A, N, is_multiset>::get_side( node_type* x, node_type*& root)
  {
    node_type* p = (*x).parent();
    if( root == x ) {
      return 0;
    }

    if( (*p).left() == x ) {
      return -1;
    }
    if( (*p).right() == x ) {
      return 1;
    }
      
    return 0; // Should never come her
  }


  /* insert_sorted - Still alot of work to be done with this routine.
 
     If the input sequence is sorted we need to do insertion in O(n) time,
     where n is the number of nodes present in the tree added with the length 
     of the input sequence. The "algorithm" is:
     We'll now generate the nodes for [x;y[ as a linked list of nodes, and 
     then call our already existing build_tree function which builds a tree
     given an array of sorted node pointers, using a devide-and-conquer 
     strategy in O(n).
     EXCEPTION SAFE. 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  template <typename I>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_sorted(I x, I y) {

    /* The tree is empty, this is similar to copy, but we don't know 
       the number of nodes in advance, we just know the sequence is sorted */

    functor_type f;
    int count = 0;
    node_type* first = 0;

    node_type* tmp;
    node_type* old = 0;

    try {

      while(x != y) {
        tmp = (*this).create_node(*x);

        if(old) {
          (*old).successor() = tmp;
          (*tmp).predecessor() = old;
        }
        if(!first) {
          first = tmp;
        }

        old = tmp;
        ++count;
        ++x;
      }
    }
    catch(...) {
      tmp = first;

      while(tmp != 0) {
        node_type* p = (*tmp).successor();
        (*this).destroy_node(p);
        tmp = p;
      }
        
      throw;
    }

    /* There is something in the tree already, so we will simply merge the lists */
    if((*this).root()) {
      node_type* undolog[count];
      unsigned int undocount = 0;

      node_type* old_tmp = (*this).leftmost();
      node_type* new_tmp = first;

      node_type* last = 0;

      first = old_tmp;

      count += node_count;

      while(old_tmp != (*this).header && new_tmp != 0) {

         bool comp;
         bool comp_equal;

         try {
           key_type new_tmp_key = f((*new_tmp).content());
           key_type old_tmp_key = f((*old_tmp).content());
           comp = (*this).comparator(new_tmp_key, old_tmp_key);
           comp_equal = !comp && !(*this).comparator(old_tmp_key, new_tmp_key);
         }
         catch(...) {
           for(int i=0; i < undocount; ++i) {
             /* Remove undolog[i] from the list */
             if((*undolog[i]).successor()) {
               (*(*undolog[i]).successor()).predecessor() = (*undolog[i]).predecessor();
             }
             if((*undolog[i]).predecessor()) {
               (*(*undolog[i]).predecessor()).successor() = (*undolog[i]).successor();
             }
             (*this).destroy_node(undolog[i]);
           }
           throw;
         }

         /* May cause an error */
         if(!is_multiset && comp_equal) {
           node_type* p = (*new_tmp).successor();
    
           if((*new_tmp).predecessor()) {
             (*(*new_tmp).predecessor()).successor() = (*new_tmp).successor();
           }
           if((*new_tmp).successor()) {
             (*(*new_tmp).successor()).predecessor() = (*new_tmp).predecessor();
           }
	       
           (*this).destroy_node(new_tmp);
           new_tmp = p;
         }
         else if(comp) {
           node_type* p = (*new_tmp).successor();

           (*new_tmp).predecessor() = (*old_tmp).predecessor();
           (*(*old_tmp).predecessor()).successor() = new_tmp;

           (*old_tmp).predecessor() = new_tmp;
           (*new_tmp).successor() = old_tmp;

           undolog[undocount] = new_tmp;
           undocount++;

           if(first == old_tmp) {
             first = new_tmp;
           }

           new_tmp = p;

         }
         else {
           old_tmp = (*old_tmp).successor();
         }

         if(old_tmp != (*this).header) {
           last = old_tmp;
         }
      }

      if(new_tmp != 0) {
        (*new_tmp).predecessor() = last;
        (*last).successor() = new_tmp;
      }

    }

    if(count > 0) {
      node_type* nodes[count];
      int i = 0;

      for(node_type* tmp = first; tmp != (*this).header && tmp != 0; tmp = (*tmp).successor()) {
        nodes[i] = tmp;
        ++i;
      }

      count = i;

      (*this).root() = (*this).build_tree(nodes, count);
      (*(*this).root()).parent() = (*this).header;
      (*this).leftmost() = (*this).tree_minimum(root());
      (*(*this).leftmost()).predecessor() = (*this).header;
      (*this).rightmost() = (*this).tree_maximum(root());
      (*(*this).rightmost()).successor() = (*this).header;
      (*(*this).header).successor() = (*this).leftmost();
      (*(*this).header).predecessor() = (*this).rightmost();
      (*this).node_count = count;
    }
  }
  
  /* erase(position) */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::size_type 
  avl_tree<K, V, F, C, A, N, is_multiset>::erase(node_type* position) {

    short link_side = (*this).get_side(position, (*(*this).header).parent());
    short link_side_new = (*this).relink_for_erase(position, 
                                                   (*(*this).header).parent(), 
                                                   (*(*this).header).left(), 
                                                   (*(*this).header).right());

    if( link_side_new != 0 ) {
      link_side = link_side_new;
    }

    (*(*position).predecessor()).successor() = (*position).successor();
    (*(*position).successor()).predecessor() = (*position).predecessor(); //this was missing, im asuming that was an error.

    node_type* z = (*position).parent();
    (*this).destroy_node(position);
    --(*this).node_count;

    // After a delete we have to rebalance
    if( (*this).root() != z ) {
      (*this).rebalance(z, (*(*this).header).parent(), -1, link_side);
    }
  }

  /* erase(k)
     EXCEPTION SAFE 
   */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::size_type 
  avl_tree<K, V, F, C, A, N, is_multiset>::erase(key_type const& k) {

    /* The upperbound for how many elements we can erase is actually the node_count */
    node_type* for_erase[node_count];
    node_type* erase_node;
    size_type count = 0;
    functor_type f;

    node_type* y = (*this).header;
    node_type* x = (*this).root();

    bool comp = true;
    erase_node = 0;

    /* We need to ensure this is actually correct, should work for multiset too.
       The try-catch thing is simply there do show that we take care of exceptions */
    try {
      while (x != 0) {

        key_type x_key = f((*x).content());
        bool comp = (*this).comparator(k, x_key);
        bool comp_equal = !(*this).comparator(x_key, k) && !comp;

        if(comp_equal) {
          erase_node = x;
          break;
        }

        if (comp) {
          x = (*x).left();
        }
        else {
          x = (*x).right();
        }
      }
    }
    catch(...) {
      throw;
    }
   
    if(!erase_node)
      return 0;

    if(!is_multiset) {
      (*this).erase(erase_node);
      return 1;
    }
 
    try {

      bool stop = false;
        
      /* we can't be sure that we found the first element in the list with the key=k, 
         so we will seek until the first */

      while(!stop) {
        key_type erase_key = f((*erase_node).content());
        key_type prev_key = f((*(*erase_node).predecessor()).content());

        if(erase_node == (*this).header) {
          stop = true;
        }

        /* erase_key != prev_key */
        else if((*this).comparator(erase_key, prev_key) || (*this).comparator(prev_key, erase_key)) {
          stop = true;
        }
        else {
          erase_node = (*erase_node).predecessor();
        }
      }
      
      stop = false;
 
      /* we'll delete all elements with equal key, so we search until k != (*k).next */

      while(!stop) {
  
        for_erase[count] = erase_node;
        count++;

        key_type erase_key = f((*erase_node).content());
        key_type next_key = f((*(*erase_node).successor()).content());

        if(erase_node == (*this).header) {
          stop = true;
        }
        /* erase_key != next_key */
        else if((*this).comparator(erase_key, next_key) || (*this).comparator(next_key, erase_key)) {
          stop = true;
        }
        else {
          erase_node = (*erase_node).successor();
        }
      }
    }
    /* we'll simply propagate the exception since nothing is done */
    catch(...) {
      throw;
    }
 
    /* now it's safe to erase the elements */
    for(int i=0; i < count; ++i) {
      (*this).erase(for_erase[i]);
    }

    return count;

  }
    
  /* relink_for_erase - Removes the node we want to delete from the tree (but doesn't delete it)
     EXCEPTION SAFE 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  short
  avl_tree<K, V, F, C, A, N, is_multiset>::relink_for_erase(node_type* z,
    node_type*& root,
    node_type*& leftmost, node_type*& rightmost) {

    node_type* x = 0;
    node_type* y = 0;
    node_type* x_parent = 0;
    short side = 0;
    y = z;
    if ((*y).left() == 0) {   // z has at most one non-null child. y == z.
      x = (*y).right();       // x might be null.
    }
    else {
      if ((*y).right() == 0) {  // z has exactly one non-null child. y == z.
        x = (*y).left();      // x is not null.
      }
      else {              // z has two non-null children.  Set y to
        y = (*y).right();     //   z's successor.  x might be null.
        while ((*y).left() != 0) {
          y = (*y).left();
        }
        x = (*y).right();
      }
    }
    node_type* new_z_parent = (*y).parent();
    if (y != z) {          // relink y in place of z.  y is z's successor
     	side = get_side( y, root); // SLY
      (*(*z).left()).parent() = y; 
      (*y).left() = (*z).left();
      if (y != (*z).right()) {
        x_parent = (*y).parent();
        if (x) {
          (*x).parent() = (*y).parent();
        }
        (*(*y).parent()).left() = x;      // y must be a left child 
        (*y).right() = (*z).right();
        (*(*z).right()).parent() = y;
      }
      else {
        x_parent = y;  
        new_z_parent = y;
      }  
      if (root == z) {
        root = y;
      }
      else if ((*(*z).parent()).left() == z) {
        (*(*z).parent()).left() = y;
      }
      else {
        (*(*z).parent()).right() = y;
      }
      (*y).parent() = (*z).parent();
      (*z).parent() = new_z_parent;
      (*y).balance() = (*z).balance();
    }
    else {                        // y == z
      x_parent = (*y).parent();
      if (x) {
        (*x).parent() = (*y).parent();   
      }
      if (root == z) {
        root = x; 
      }
      else { 
        if ((*(*z).parent()).left() == z) {
          (*(*z).parent()).left() = x;
        }
        else {
          (*(*z).parent()).right() = x;
        }
      }
      if (leftmost == z) {
        if ((*z).right() == 0) {       // left[z] must be null also
          leftmost = (*z).parent();
        }
        // makes leftmost == m_header if z == root
        else {
          leftmost = (*this).tree_minimum(x);
        }
      }
      if (rightmost == z) {
        if ((*z).left() == 0) {         // right[z] must be null also
          rightmost = (*z).parent();  
        }
        // makes rightmost == m_header if z == root
        else {                      // x == left[z]
          rightmost = (*this).tree_maximum(x);
        }
      }
    }

    return side;

  }
 
  /* find_node - Using the function z to search the tree. 
     EXCEPTION SAFE 
  */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  template <typename Z>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::find_node(const key_type& k, Z z) const
  {
    node_type* y = (*this).header; /* Last node which is not less than k. */
    node_type* x = (*this).root(); /* Current node. */
    while (x != 0) {
      if (z((*this), x, k)) {
        y = x;
        x = (*x).left();
      }
      else {
        x = (*x).right();
      }
    }
    return y;
  }

  /* lower_bound(k)
     EXCEPTION SAFE 
   */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::lower_bound(const key_type& k) const
  {
    return (*this).find_node(k, f_lowerbound<K, V, F, C, A, N, is_multiset>());
  }

  /* upper_bound(k)
     EXCEPTION SAFE 
   */
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::upper_bound(const key_type& k) const
  {
    return (*this).find_node(k, f_upperbound<K, V, F, C, A, N, is_multiset>());
  }

  /* swap */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::swap(avl_tree<K, V, F, C, A, N, is_multiset>& t) {
    std::swap((*this).header, t.header);
    std::swap((*this).node_count, t.node_count);
    std::swap((*this).comparator, t.comparator);
    std::swap((*this).allocator, t.allocator);
  }

  /* casting operator - it makes it possible to print out the data structure 
     (with balance information, ..) 
   */

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  avl_tree<K, V, F, C, A, N, is_multiset>::operator std::string() const {
    if((*this).root() == 0) { 
      return std::string("The container is empty!"); 
    }

    return std::string(*(*this).root());  
  }

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  std::pair<typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*, bool>
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(const value_type& v, cphstl::bool2type<false>) {

    functor_type f;

    node_type* y = (*this).header;
    node_type* x = (*this).root();

    node_type* pred = (*this).header;
    node_type* succ = (*this).header;

    key_type k = f(v);

    bool comp = true;

    while (x != 0) {
      y = x;
 
      key_type x_key = f((*x).content());
      comp = (*this).comparator(k, x_key);
      if (comp) {
        if(!(*this).comparator(x_key, k)) {
          succ = x;
        }
        x = (*x).left();
      }
      else {
        if((*this).comparator(x_key, k)) {
          pred = x;
        }
        x = (*x).right();
      }
    }

    node_type* j = y;
    if (comp) {
      if (j == (*this).begin()) {
        return std::pair<node_type*,bool>((*this)._insert(x, y, v, pred, succ), true);
      }
      else {
        j = (*j).predecessor();
      }
    }

    key_type j_key = f((*j).content());

    if ((*this).comparator(j_key, k)) {
      return std::pair<node_type*,bool>((*this)._insert(x, y, v, pred, succ), true);
    }
    return std::pair<node_type*,bool>(j, false);

  }

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(const value_type& v, cphstl::bool2type<true>) {
    functor_type f;

    node_type* y = (*this).header;
    node_type* x = (*this).root();

    node_type* pred = (*this).header;
    node_type* succ = (*this).header;

    bool comp = true;
    key_type k = f(v);

    while (x != 0) {
      y = x;
      key_type x_key = f((*x).content());
   
      comp = (*this).comparator(k, x_key);
      if (comp) {
        if(!(*this).comparator(x_key,k) || (!(*this).comparator(x_key, k) && !(*this).comparator(k, x_key))) {
          succ = x;
        }
        x = (*x).left();
      }
      else {
        if((*this).comparator(x_key, k) || (!(*this).comparator(x_key, k) && !(*this).comparator(k, x_key))) {
          pred = x;
        }
        x = (*x).right();
      }
    }

    return (*this)._insert(x, y, v, pred, succ);
  }

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(node_type* position, const value_type& v, cphstl::bool2type<false>) {

    functor_type f;

    key_type k = f(v);
    key_type position_key = f((*position).content());

    if (position == (*(*this).header).left()) { // begin()
      if ((*this).size() > 0 && (*this).comparator(k, position_key)) {
        return (*this)._insert(position, position, v, (*position).predecessor(), position);
      }
      // first argument just needs to be non-null 
      else {
        return (*this).insert(v).first;
      }
    } else if (position == (*this).header) { // end()
      key_type rightmost_key = f((*(*this).rightmost()).content());
      if ((*this).comparator(rightmost_key, k))
        return (*this)._insert(0, (*this).rightmost(), v, (*position).predecessor(), (*this).header);
      else
        return (*this).insert(v).first;
    } else {
      node_type* before = position;
      before = (*before).predecessor();
      key_type before_key = f((*before).content());
      /* There may be an error here too */
      if ((*this).comparator(before_key, k) && (*this).comparator(k, position_key)) {
        if ((*before).right() == 0) {
          return (*this)._insert(0, before, v, (*before).predecessor(), (*before).successor()); 
	}
        else {
          return (*this)._insert(position, position, v, (*position).predecessor(), (*position).successor());
          // first argument just needs to be non-null 
	}
      } else {
        return (*this).insert(v).first;
      }
    }

  }

  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  typename avl_tree<K, V, F, C, A, N, is_multiset>::node_type*
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(node_type* position, const value_type& v, cphstl::bool2type<true>) {

    functor_type f;
    key_type k = f(v);
    key_type position_key = f((*position).content());

    if (position == (*(*this).header).left()) { // begin()
      if (size() > 0 && (*this).comparator(k, position_key))
        return _insert(position, position, v, (*position).predecessor(), (*position).successor());
      // first argument just needs to be non-null 
      else
        return (*this).insert(v);
    } else if (position == (*this).header) { // end()
      key_type rightmost_key = f((*(*this).rightmost()).content());
      if (!(*this).comparator(k, rightmost_key))
        return (*this)._insert(0, (*this).rightmost(), v, (*position).predecessor(), (*this).header);
      else
        return (*this).insert(v);
    } else {
      node_type* before = position;
      before = (*before).predecessor();
      key_type before_key = f((*before).content());

      if (!(*this).comparator(k, before_key) && !(*this).comparator(position_key, k)) {
        if ((*before).right() == 0)
          return (*this)._insert(0, before, v, (*before).predecessor(), (*before).successor()); 
        else
          return (*this)._insert(position, position, v, (*position).predecessor(), (*position).successor());
          // first argument just needs to be non-null 
      } else
        return (*this).insert(v);
    }

  }


  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  template <typename I>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(I x, I y, cphstl::bool2type<false>) {
    int count = 0;
    for(I it = x; it != y; ++it)
      ++count;

    node_type* undolog[count];
    count = 0;
    try {
      for(I it = x; it != y; ++it) {
        std::pair<node_type*, bool> pair;
        pair = (*this).insert(*it);
        if(pair.second) {
          undolog[count] = pair.first;
          ++count;
        }
      }
    }
    catch (...) {
      for (int i = 0;i<=count;++i) {
        (*this).erase(undolog[i]);
      } 
	    throw;
    }
  }
 
  template <typename K, typename V, typename F, typename C, typename A, typename N, bool is_multiset>
  template <typename I>
  void
  avl_tree<K, V, F, C, A, N, is_multiset>::insert_dispatch(I x, I y, cphstl::bool2type<true>) {
    int count = 0;

    for(I it = x; it != y; ++it)
      ++count;

    node_type* undolog[count];

    count = 0;
    try {
      for(I it = x; it != y; ++it) {
        undolog[count] = (*this).insert(*it);
      }
    }
    catch (...) {
      for (int i = 0;i<=count;++i) {
        (*this).erase(undolog[i]);
      } 
      throw;
    }

  }
}
