/*

  Header file for the AVL implementation

  Authors: Bo Simonsen, May 2008, December 2008

*/

#ifndef CPHSTL_AVL_HPP
#define CPHSTL_AVL_HPP

#include <functional>
#include <algorithm>
#include <memory>

#include "assert.h++"
#include "type.h++"
#include "is_sorted.c++" // Defines cphstl::is_sorted generic algorithm
#include "avl_tree_node.h++"
#include "avl_comp.h++"

#include <iostream>

namespace cphstl {

  /* Forward declarations */
  template <typename K,typename V,typename F, typename C, typename A,typename N, bool is_multiset>
  class f_lowerbound;
  template <typename K,typename V,typename F, typename C, typename A,typename N, bool is_multiset>
  class f_upperbound;

  template <
    typename K, 
    typename V = K,
    typename F = cphstl::unnamed::identity<V>,
    typename C = std::less<K>, 
    typename A = std::allocator<V>,
    typename N = cphstl::avl_tree_node<V>,
    bool is_multiset = false
  >
  class avl_tree
  {
    friend class f_upperbound<K,V,F,C,A,N,is_multiset>;       
    friend class f_lowerbound<K,V,F,C,A,N,is_multiset>;        

  public:
    typedef K key_type;
    typedef V value_type;
    typedef C key_compare;
    typedef N node_type; 
    typedef F functor_type;          

    typedef typename cphstl::if_then_else<cphstl::types<K, V>::are_same, C, _value_compare<K, V, C> >::type value_compare;
    typedef typename A::template rebind<node_type>::other allocator_type;

    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef value_type* pointer;
    typedef value_type const* const_pointer;
    typedef value_type& reference;
    typedef value_type const& const_reference;      	

  private:
    key_compare comparator;
    allocator_type allocator; 
    node_type*     header;
    size_type node_count;

  public:
    /* Constructors */
    explicit avl_tree(const key_compare& k, const allocator_type& a);
    avl_tree(const avl_tree& x);
    ~avl_tree();

    /* Operators */
    avl_tree& operator=(const avl_tree& x);

    /* Size accessors */	   
    size_type size() const;
    size_type max_size() const;

    /* Policy accessors */
    allocator_type get_allocator() const;
    key_compare key_comp() const;
    value_compare value_comp() const;

    /* Accessing and manipulating methods */
    node_type* lower_bound(const key_type& k) const;      
    node_type* upper_bound(const key_type& k) const;      
      
    node_type* begin() const;
    node_type* end() const;

    typename cphstl::if_then_else<is_multiset, 
                           node_type*, 
                           std::pair<node_type*, bool> 
                          >::type
    insert(value_type const& v);

    node_type* insert(node_type* position, value_type const& v);

    template <typename I>
    void insert(I x, I y);

    size_type erase(node_type* position);
    size_type erase(key_type const& position);      
 
    void swap(avl_tree& t);
    operator std::string() const;

  protected:
    node_type* make_node();
    void free_node(node_type* p);

    node_type* create_node(const value_type& x);

    void destroy_node(node_type* p);
    void destroy_tree(node_type* x);

    void construct(value_type* p, value_type const& v) const;
    void destroy(value_type* p)   const;
    void init();

    short get_side( node_type* x, node_type*& root);
    void rebalance(node_type* x, node_type*& root, int cmd, short side);
    void rotate_left( node_type* x, node_type*& root);
    void rotate_right( node_type* x, node_type*& root);

    short relink_for_erase(node_type* z, node_type*& root, node_type*& leftmost, node_type*& rightmost);

    node_type* copy_nodes(node_type* begin, node_type* end, int node_count);
    void copy(const avl_tree& x);
    node_type* build_tree(node_type* nodes[], int node_count);

    node_type*& root()      const;
    node_type*& leftmost()  const;
    node_type*& rightmost() const;

    node_type* tree_maximum(node_type* x) const;
    node_type* tree_minimum(node_type* x) const;

    node_type* _insert(node_type* x, node_type* y, const value_type& v, node_type* pred, node_type* succ);

    /* template parameter F would be appropiate, but it is already in use, so Z is randomly selected */
    template <typename Z>
    node_type* find_node(const key_type& k, Z z) const;

    /* Dispatching methods */
    std::pair<node_type*, bool> insert_dispatch(const value_type& v, cphstl::bool2type<false>);
    node_type* insert_dispatch(const value_type& v, cphstl::bool2type<true>);

    node_type* insert_dispatch(node_type* position, const value_type& v, cphstl::bool2type<false>);
    node_type* insert_dispatch(node_type* position, const value_type& v, cphstl::bool2type<true>);
   
    template <typename I>
    void insert_dispatch(I x, I y, cphstl::bool2type<false>);
    template <typename I>
    void insert_dispatch(I x, I y, cphstl::bool2type<true>);
  
    template <typename I>
    void insert_sorted(I x, I y);
  };
} // namespace cphstl

#include "avl_func.h++"
#include "avl_tree.c++"

#endif // CPHSTL_AVL_HPP

