/*
  An implementation of cphstl::doubly_linked_list

  Status: remove_if, unique, merge and sort not implemented

  Author: Jyrki Katajainen, April 2008
*/

#include <cassert> // defines assert
#include <algorithm> // defines std::swap

namespace cphstl {

  // default constructor

  template <typename V, typename A, typename N>
  doubly_linked_list<V, A, N>::doubly_linked_list(A const& a) 
    : node_allocator(a), n(0) {
    N* s = create_sentinel();
    (*this).sentinel = s;
  }

  // parametrized constructors

  template <typename V, typename A, typename N>
  doubly_linked_list<V, A, N>::doubly_linked_list(size_type k, V const& v, A const& a) 
    : node_allocator(a), n(0) { 
    N* s = (*this).create_sentinel();
    try {
      (*this).insert(s, static_cast<size_type>(k), v);
    }
    catch (...) {
      (*this).destroy_sentinel(s);
      throw;
    }
    (*this).sentinel = s;      
  }

  template <typename V, typename A, typename N>
  template <typename I>
  doubly_linked_list<V, A, N>::doubly_linked_list(I p, I q, A const& a)
    : node_allocator(a), n(0) { 
    N* s = (*this).create_sentinel();
    try {
      for ( ; p != q; ++p) {
	(void)(*this).insert(s, *p);
      }
    }
    catch (...) {
      (*this).erase_range((*s).successor(), s);
      (*this).destroy_sentinel(s);
      throw;
    }
    (*this).sentinel = s;
  }

  // copy constructor

  template <typename V, typename A, typename N>
  doubly_linked_list<V, A, N>::doubly_linked_list(doubly_linked_list const& x)
    : node_allocator(x.node_allocator), n(0) {
    N* s = create_sentinel();
    try {
      (*this).insert_range(s, x.begin(), x.end());
    }
    catch(...) {
      (*this).destroy_sentinel(s);
      throw;
    }
    (*this).sentinel = s;
  }

  // destructor

  template <typename V, typename A, typename N>
  doubly_linked_list<V, A, N>::~doubly_linked_list() { 
    (*this).erase_range((*this).begin(), (*this).end());
    (*this).destroy_sentinel((*this).end());
  }

  // operator=

  template <typename V, typename A, typename N>
  doubly_linked_list<V, A, N>& 
  doubly_linked_list<V, A, N>::operator=(doubly_linked_list const& x) {
    if (this == &x) {
      return *this;
    }
    N* s = create_sentinel();
    try {
      (*this).insert_range(s, x.begin(), x.end());
      try {
	(*this).node_allocator = x.node_allocator;
      }
      catch(...) {
	(*this).erase_range((*s).successor(), s);
	throw;
      }
    }
    catch(...) {
      (*this).destroy_sentinel(s);
      throw;
    }
    (*this).erase_range((*this).begin(), (*this).end());
    (*this).destroy_sentinel((*this).sentinel);
    (*this).sentinel = s;
    return *this;
  }

  // begin

  template <typename V, typename A, typename N>
  N*
  doubly_linked_list<V, A, N>::begin() {
    return (*(*this).sentinel).successor();
  }

  template <typename V, typename A, typename N>
  N const*
  doubly_linked_list<V, A, N>::begin() const {
    return (*(*this).sentinel).successor();
  }

  // end

  template <typename V, typename A, typename N>
  N*
  doubly_linked_list<V, A, N>::end() {
    return (*this).sentinel;
  }

  template <typename V, typename A, typename N>
  N const*
  doubly_linked_list<V, A, N>::end() const {
    return (*this).sentinel;
  }

  // get_allocator

  template <typename V, typename A, typename N>
  A
  doubly_linked_list<V, A, N>::get_allocator() const {
    return A((*this).node_allocator);
  }

  // size

  template <typename V, typename A, typename N>
  typename doubly_linked_list<V, A, N>::size_type 
  doubly_linked_list<V, A, N>::size() const {
    return (*this).n;
  }

  // max_size

  template <typename V, typename A, typename N>
  typename doubly_linked_list<V, A, N>::size_type 
  doubly_linked_list<V, A, N>::max_size() const {
    return node_allocator.max_size(); // available memory in nodes
  }

  // single-element insert

  template <typename V, typename A, typename N>
  N*
  doubly_linked_list<V, A, N>::insert(N* r, V const& v) {
    N* q = (*this).create_node(v);
    N* p = (*r).predecessor();
    arc(p, q);
    arc(q, r);
    (*this).n +=1;
    return q;
  }

  // multiple-element inserts

  template <typename V, typename A, typename N>
  void
  doubly_linked_list<V, A, N>::insert(N* q, size_type m, V const& v) {
    N* const p = (*q).predecessor();
    try {
      for (size_type i = size_type(0); i < m; ++i) {
	(void)(*this).insert(q, v);
      }
    }
    catch (...) {
      (*this).erase_range((*p).successor(), q);
      throw;
    }
  }

  template <typename V, typename A, typename N>
  template <typename I>
  void
  doubly_linked_list<V, A, N>::insert(N* q, I r, I s) {
    N* const p = (*q).predecessor();
    try {
      for (; r != s; ++r) {
	(void)(*this).insert(q, *r);
      }
    }
    catch (...) {
      (*this).erase_range((*p).successor(), q);
      throw;
    }
  }

  // single-element erase

  template <typename V, typename A, typename N>
  N*
  doubly_linked_list<V, A, N>::erase(N* q) {
    assert(q != (*this).sentinel);
    N* p = (*q).predecessor();
    N* r = (*q).successor();
    arc(p, r);
    (*this).destroy_node(q);
    (*this).n -= 1;
    return r;
  }

  // swap

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::swap(doubly_linked_list<V, A, N>& x) {
    assert((*this).node_allocator == x.node_allocator);
    std::swap((*this).sentinel, x.sentinel);
    std::swap((*this).n, x.n);
  }

  // splice

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::splice(N* s, doubly_linked_list<V, A, N>& x) {
    if (x.n == size_type(0)) {
      return;
    }
    assert((*this).node_allocator == x.node_allocator);
    N* p = (*s).predecessor();
    N* q = x.begin();
    N* t = x.end();
    N* r = (*t).predecessor();
    arc(p, q);
    arc(r, s);
    arc(t, t);
    (*this).n += x.n;
    x.n = size_type(0);
  }

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::splice(N* r, doubly_linked_list<V, A, N>& x, N* q) {
    if (r == q || r == (*q).successor()) {
      return;
    }
    assert((*this).node_allocator == x.node_allocator);
    N* p = (*r).predecessor();
    N* s = (*q).predecessor();
    N* t = (*q).successor();
    arc(p, q);
    arc(q, r);
    arc(s, t);
    (*this).n += 1;
    x.n -= 1;
  }

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::splice(N* t, doubly_linked_list<V, A, N>& x, N* q, N* s) {
    if (q == s) {
      return;
    }
    assert((*this).node_allocator == x.node_allocator);
    size_type m = size_type(0);
    if (this != &x) {
      m = (*this).distance(q, s);
    }
    N* o = (*t).predecessor();
    N* p = (*q).predecessor();
    N* r = (*s).predecessor();
    arc(o, q);
    arc(r, t);
    arc(p, s);
    (*this).n += m;
    x.n -= m;
  }

  // remove_if

  template <typename V, typename A, typename N>
  template <typename P>
  void 
  doubly_linked_list<V, A, N>::remove_if(P unary_predicate) {
    // not implemented
  }

  // unique

  template <typename V, typename A, typename N>
  template <typename P>
  void 
  doubly_linked_list<V, A, N>::unique(P binary_predicate) {
    // not implemented
  }

  // merge

  template <typename V, typename A, typename N>
  template <typename C>
  void 
  doubly_linked_list<V, A, N>::merge(doubly_linked_list<V, A, N>& x, C c) {
    // not implemented
  }

  // sort

  template <typename V, typename A, typename N>
  template <typename C>
  void 
  doubly_linked_list<V, A, N>::sort(C c) {
    // not implemented
  }

  // reverse

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::reverse() {
    N* p = (*this).sentinel;
    do {
      N* q = (*p).successor();
      std::swap((*p).predecessor(), (*p).successor());
      p = q;
    } while (p != (*this).sentinel);
  }

  // arc

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::arc(N* p, N* q) {
    (*p).successor() = q;
    (*q).predecessor() = p;
  }

  // create_node

  template <typename V, typename A, typename N>
  N*
  doubly_linked_list<V, A, N>::create_node(V const& value) {
    N* p = node_allocator.allocate(1);
    new (p) N(value);
    return p;
  }

  // create_sentinel

  template <typename V, typename A, typename N>
  N* 
  doubly_linked_list<V, A, N>::create_sentinel() {
    N* s = node_allocator.allocate(1);
    arc(s, s);
    return s;
  }

  // destroy_node

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::destroy_node(N* p) {
    (*p).~N();
    node_allocator.deallocate(p, 1);
  }    

  // destroy_sentinel

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::destroy_sentinel(N* p) {
    node_allocator.deallocate(p, 1);
  }

  // distance

  template <typename V, typename A, typename N>
  typename doubly_linked_list<V, A, N>::size_type
  doubly_linked_list<V, A, N>::distance(N* p, N* r) {
    size_type m = size_type(0);
    for (N* q = p; q != r; q = (*q).successor()) {
      ++m;
    }
    return m;
  }

  // erase_range

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::erase_range(N* q, N* r) {
    while (q != r) {
      q = (*this).erase(q);
    }
  }

  // insert_range

  template <typename V, typename A, typename N>
  void 
  doubly_linked_list<V, A, N>::insert_range(N* s, N const* q, N const* r) {
    N* p = (*s).predecessor();
    try {
      for (; q != r; q = (*q).successor()) {
	(void)(*this).insert(s, (*q).content());
      }
    }
    catch(...) {
      (*this).erase_range((*p).successor(), s);
      throw;
    }
  }

}

