/*
  A doubly-linked list is a simple realization of a list.

  Author: Jyrki Katajainen, April 2008
*/

#ifndef __CPHSTL_DOUBLY_LINKED_LIST__
#define __CPHSTL_DOUBLY_LINKED_LIST__

#include <memory> // defines std::allocator
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <functional> // defines std::less

#include "list_node.h++" // defines cphstl::list_node
#include "node_iterator.h++" // defines cphstl::node_iterator

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>,
    typename N = cphstl::list_node<V>
  >
  class doubly_linked_list {
  public:

    // types
    
    typedef V value_type;
    typedef A allocator_type;
    typedef N node_type;
    typedef std::size_t size_type;
    typedef cphstl::node_iterator<N, false> iterator;
    typedef cphstl::node_iterator<N, true> const_iterator;
    typedef N* concrete_iterator;
    typedef N const* concrete_const_iterator;

    // structors
  
    explicit doubly_linked_list(A const& = A());
    explicit doubly_linked_list(size_type, V const& = V(), A const& = A());
  
    template <typename I>
    doubly_linked_list(I, I, A const& = A());
  
    doubly_linked_list(doubly_linked_list const&);
    ~doubly_linked_list();
    doubly_linked_list& operator=(doubly_linked_list const&);

    // iterators

    N* begin();
    N const* begin() const;
    N* end();
    N const* end() const;

    // accessors

    A get_allocator() const;
    size_type size() const;
    size_type max_size() const;

    // modifiers

    N* insert(N*, V const&);
    void insert(N*, size_type, V const&);

    template <typename I>
    void insert(N*, I, I);

    N* erase(N*);
    void swap(doubly_linked_list&);

    // list operations

    void splice(N*, doubly_linked_list&);
    void splice(N*, doubly_linked_list&, N*);
    void splice(N*, doubly_linked_list&, N*, N*);

    template <typename P> 
    void remove_if(P);

    template <typename P>
    void unique(P);

    template <typename C> 
    void merge(doubly_linked_list&, C);

    template <typename C> 
    void sort(C);

    void reverse();
  
  protected:

    typedef typename A::template rebind<N>::other node_allocator_type;

    node_allocator_type node_allocator;
    N* sentinel; 
    size_type n;

  private:

    void arc(N*, N*);
    N* create_node(V const&);
    N* create_sentinel();
    void destroy_node(N*);    
    void destroy_sentinel(N*);
    size_type distance(N*, N*);
    void erase_range(N*, N*);
    void insert_range(N*, N const*, N const*);
    
  };

}

#include "doubly_linked_list.c++" // implements cphstl::doubly_linked_list

#endif
