/*
  An implementation of cphstl::dynamic_array

  Authors: Tina A. G. Andersen, Jyrki Katajainen, Mads D. Kristensen,
  Ulrik Schou Jørgensen and Claus Ullerlund, April 2008
  Bo Simonsen, May 2008

  Important: This class cannot be used standalone. This class is a
  realization to be used with the bridge-class vector.

  The design pattern used is called the "bridge pattern":
  
  client <--> bridge <--> realization

  The advantage of this is you can change the realization without
  confusing the client.
*/

#include <cassert> // defines assert
#include <algorithm> // defines std::swap
#include <cstdio>

namespace cphstl {

  // constructor: create an empty array

  template <typename V, typename A, typename E>
  dynamic_array<V, A, E>::dynamic_array(A const& a) 
    : current_capacity(1), current_size(0), allocator(a) {
    array_p = allocator.allocate(current_capacity);
  }
  // copy constructor

  template <typename V, typename A, typename E>
  dynamic_array<V, A, E>::dynamic_array(dynamic_array const& x)
    : current_capacity(x.current_capacity), current_size(x.current_size), 
      allocator(x.allocator) {
    array_p = allocator.allocate(current_capacity);
    pointer target = array_p;
    const_pointer source_end = x.array_p + x.current_capacity;
    for (const_pointer source = x.array_p; source < source_end; ++source) {
      new (target) E(*source);
      ++target;
    }
  }
      
  // destructor

  template <typename V, typename A, typename E>
  dynamic_array<V, A, E>::~dynamic_array() {
    pointer end_p = array_p + current_size;
    for (pointer p = array_p; p < end_p ; ++p) {
      //http://www.research.att.com/~bs/bs_faq2.html
      (*p).~E();
    }
    allocator.deallocate(array_p, current_size);
  }

  // = operator overloading

  template <typename V, typename A, typename E>
  dynamic_array<V,A>&
  dynamic_array<V, A, E>::operator=(const dynamic_array<V,A> &x){
    current_capacity = x.current_capacity;
    current_size     = x.current_size;
    array_p = allocator.allocate(current_capacity);
    pointer target = array_p;
    const_pointer source_end = x.array_p + x.current_capacity;
    for (const_pointer source = x.array_p; source < source_end; ++source) {
      new (target) E(*source);
      ++target;
    }
    return (*this);
  }

  // begin: return a random-access iterator referring to the first element

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::concrete_iterator
  dynamic_array<V, A, E>::begin() {
    return &array_p[0];
  }

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::concrete_const_iterator
  dynamic_array<V, A, E>::begin() const {
    return &array_p[0];
  }
  
  // get_allocator: return the allocator

  template <typename V, typename A, typename E>
  A
  dynamic_array<V, A, E>::get_allocator() const {
    return A(allocator);
  }

  // size: return the number of elements stored

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::size_type
  dynamic_array<V, A, E>::size() const {
    return current_size;
  }

  // max_size: return the maximum number of elements possible

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::size_type
  dynamic_array<V, A, E>::max_size() const {
    return current_capacity + allocator.max_size();
  }

  // capacity: return the maximum number of elements without reallocation

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::size_type 
  dynamic_array<V, A, E>::capacity() const {
    return current_capacity;
  }

  // operator[]: return a const reference to the element with index s

  template <typename V, typename A, typename E>
  V const&
  dynamic_array<V, A, E>::operator[](size_type s) const {
    return array_p[s].content();
  }

  // operator[]: return a reference to the element with index s

  template <typename V, typename A, typename E>
  V&
  dynamic_array<V, A, E>::operator[](size_type s) {
    return array_p[s].content();
  }

  // reserve: enlarge the capacity, if not large enough
  // Correct: variable names misleading; variables are pointers not iterators

  template <typename V, typename A, typename E>
  void
  dynamic_array<V, A, E>::reserve(size_type s) {
    if (s <= current_capacity){
      return;
    }
    assert(current_size < current_capacity+1);
    // Allocate new space for the elements.
    pointer new_array_iter = allocator.allocate(s);
    
    // Move the objects from the old array into the new one. While we do
    // this we also destroy the old objects.
    pointer old_end_iter = array_p + current_size;
    pointer iter_new = new_array_iter;
    pointer iter_old = array_p;
    while (iter_old < old_end_iter) {
      new (iter_new) E(*iter_old);
      iter_new++;
      (*iter_old).~E();
      iter_old++;
    }
	  
    // Deallocate the old space and swap the pointers.
    allocator.deallocate(array_p, current_capacity);
    array_p = new_array_iter;
    current_capacity = s;

    assert(current_size < current_capacity + 1);
  }

  // insert: insert a copy of v before position p; return its position
  // Correct: variable names misleading; variables are pointers not iterators

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::concrete_iterator
  dynamic_array<V, A, E>::insert(concrete_iterator p, value_type const& v) {
    
    // This has been split up into two parts, one where resizing is
    // needed and one that doesn't need resizing. This way less
    // element moving is done when a resize is needed.
    assert(current_size < current_capacity+1);
    if (current_size == current_capacity - 1) {
      
      // Increase the capacity of the array.
      pointer new_array_p = allocator.allocate(current_capacity * 2);
      
      // First we copy the elements in front of pos into the new array.
      pointer p_new = new_array_p;
      pointer p_old = array_p;
      while (p_old < p) {
        new (p_new) E(*p_old);
        p_new++;
        (*p_old).~E();
        p_old++;
      }
      
      // Then we insert the new element and copy the rest of the
      // elements into the new array.
      p = p_new;
      new (p_new) E(v);
      p_new++;
      pointer old_end_p = array_p + current_size;
      while (p_old < old_end_p) {
        new (p_new) E(*p_old);
        p_new++;
        (*p_old).~E();
        p_old++;
      } 
      
      // The last thing to do is to deallocate the old array
      // and swap the pointers.
      allocator.deallocate(array_p, current_capacity);
      current_capacity *= 2;
      array_p = new_array_p;
    }
    else {
      
      // In this case we move all the elements from pos and forward
      // one position to the right.
      pointer p_ = array_p + current_size;
      while (p_ > p) {
        /* This is not initialized! */
        if(p_ != array_p + current_size)
          (*p_).~E();
        new (p_) E(*(p_-1));
        p_--;
      }
      
      // And insert the element in the correct position.
      (*p).~E();
      new (p) E(v);
    }
    
    current_size += 1;
    assert(current_size <= current_capacity);
    return p;
  }
  
  // insert: insert s copies of v before position p; return nothing
  // Correct: This version has a quadratic running time in the worst case!

  template <typename V, typename A, typename E>
  void
  dynamic_array<V, A, E>::insert(concrete_iterator p, size_type s, value_type const& v) {

    assert(current_size <= current_capacity); 
    // set the target_array og to the original 
    pointer target_p = array_p;
    //if we need to resize : calculate new target and resize properly
    if (current_size + s >= current_capacity) {
      size_type new_capacity = current_capacity;
      while(current_size + s >= new_capacity){
        new_capacity*=2;
      }
      //set new target and capacity;
      target_p = allocator.allocate(new_capacity);
      current_capacity = new_capacity;

      // First we copy the elements in front of pos into the new array.
      pointer p_new = target_p;
      pointer p_old = array_p;
      while (p_old < p) {
        new (p_new) E(*p_old);
        p_new++;
        (*p_old).~E();
        p_old++;
      }
      //Now we insert s elements in front
      for(size_type i = 0; i < s;++i){
        new (p_new) E(v);
        p_new++;
      }
      //copy the rest in and put them in front
      while (p_old< array_p + current_size){
        new (p_new) E(*p_old);
        p_new++;
        (*p_old).~E();
        p_old++;
      } 
      allocator.deallocate(array_p, current_capacity);
      current_capacity = new_capacity;
      array_p = target_p;
    }
    else {
      //If we dont need to increase capity;
      //First copy the old element forward to their new position
      pointer p_new = array_p + current_size + s;
      pointer p_end = p + s ;

      while (p_new >= p_end) {
        (*p_new).~E();
        new (p_new) E(*(p_new-s));
        --p_new;
      }

      //Now create the new entries
      for(size_type i = 0; i < s ; ++i){
        (*p_new).~E();
        new(p_new) E(v);
        --p_new;
      }
    }
   current_size += s;

    assert(current_size <= current_capacity); 
  }

  // insert: insert a copy of elements in the range [q,r) before position p
  // Correct: This version has a quadratic running time in the worst case!

  template <typename V, typename A, typename E>
  template <typename I>
  void
  dynamic_array<V, A, E>::insert(concrete_iterator p, I q, I r) {    
    int index = p - array_p;

    // Find the number of elements to insert 
    I iter = q;
    size_type count = 0;
    while(iter != r) {
      ++count;
      ++iter;
    }

    // We let this insert function do the work for us
    (*this).insert(p, count, V());

    iter = q;
    size_type i = index;

    // Destroy the old elements and insert the new ones 
    while(iter != r) {
      //assert((*this)[i] == V());
      // using the [] operator
      (*this)[i].~V();
      (*this)[i] = *iter;
      ++iter;
      ++i;
    }

  }

  // erase: remove the element at position p; return the position of the next element
  // Correct: variable names misleading; variables are pointers not iterators

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::concrete_iterator
  dynamic_array<V, A, E>::erase(concrete_iterator p) {
    assert(current_size <= current_capacity);
    pointer end_p = array_p + current_size - 1;
    pointer p_ = p;
    (*p_).~E();
    while (p_ < end_p) {
      new (p_) E(*(p_ + 1));
      ++p_;
      (*p_).~E();
    }
    --current_size;
    assert(current_size <= current_capacity);
    return p;
  }

  // erase: remove all elements in the range [p,q); return q

  template <typename V, typename A, typename E>
  typename dynamic_array<V, A, E>::concrete_iterator
  dynamic_array<V, A, E>::erase(concrete_iterator p, concrete_iterator q) {
    pointer end_p = array_p + current_size;
    pointer p_ = p;
    pointer q_ = q;
    int range = q_ - p_; 
    pointer oldp = p_;
    while (p_ < q_){
      (*p_).~E();
      ++p_;
    }
    p_ = oldp;
    while (p_ < end_p) {
      new (p_) E(*(p_ + range));
      (*(p_ + range)).~E();
      ++p_;
    }
    current_size = current_size - range;
    return q;
  }

  // swap: swap the data of the calling dynamic array and r

  template <typename V, typename A, typename E>
  void 
  dynamic_array<V, A, E>::swap(dynamic_array<V, A, E>& r) {
    assert(allocator == r.allocator);
    std::swap(array_p, r.array_p);
    std::swap(current_size, r.current_size);
    std::swap(current_capacity, r.current_capacity);
  }

}
