
/*
  A dynamic array is a container that doubles its capacity when it
  becomes full, and halves its capacity when it becomes only one-forth
  full.

  Authors: Tina A. G. Andersen, Jyrki Katajainen, Ulrik Schou Joergensen and Claus Ullerlund, Bo Simonsen, 
  April 2008


  Important : This class cannot be used standalone. This class is a _realisation_ to be used with the _bridge-class_ vector(../../Vector).

  The design pattern is called a "bridge pattern":
  
  client <--> bridge <--> realisation

  The advantage of this, is you can change the realisation without confusing the client.

*/

#ifndef __CPHSTL_DYNAMIC_ARRAY__
#define __CPHSTL_DYNAMIC_ARRAY__

#include <memory> // defines std::allocator
#include <stdexcept> // defines std::out_of_range
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <utility> // defines std::pair

#include "entry_iterator.h++" // defines cphstl::entry_iterator

namespace cphstl {

  template <typename V, typename A>
  class dynamic_array_entry;

  template <
    typename V, 
    typename A = std::allocator<V>,
    typename E = dynamic_array_entry<V, A>
  >
  class dynamic_array {
    
  public:
    
    // types

    typedef V value_type;
    typedef A allocator_type;
    typedef E* pointer;
    typedef E const* const_pointer;
    typedef E entry;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;
    typedef E* concrete_iterator;
    typedef const E* concrete_const_iterator;

    // structors

    explicit dynamic_array(A const& = A());
    explicit dynamic_array(size_type, V const& = V(), A const& = A());
    dynamic_array(dynamic_array<V, A, E> const&);
    ~dynamic_array();
    
    // iterators

    concrete_iterator begin();
    concrete_const_iterator begin() const;
    
    // accessors

    A get_allocator() const;
    size_type size() const;
    size_type max_size() const;
    size_type capacity() const;
    V const& operator[](size_type) const;

    dynamic_array<V,A>& operator=(const dynamic_array<V,A>&);
    
    // modifiers

    V& operator[](size_type);
    void reserve(size_type);
    concrete_iterator insert(concrete_iterator, V const&);
    void insert(concrete_iterator, size_type, V const&);
    
    template <typename I>
    void insert(concrete_iterator, I, I);
    
    concrete_iterator erase(concrete_iterator);
    concrete_iterator erase(concrete_iterator, concrete_iterator);
    void swap(dynamic_array<V, A, E>&);
    
  protected:

    typedef typename A::template rebind<E>::other value_allocator_type;   
    
    size_type current_capacity;
    size_type current_size;
    pointer array_p;
    value_allocator_type allocator;
    
  };
  
}

#include "dynamic_array_entry.h++"
#include "dynamic_array.c++" //implements cphstl::dynamic_array

#endif
