/*
  A rank-based iterator is just a (pointer, index) pair where the
  pointer points to the data structure containing the cell referred to
  and the index is the rank of that cell in the sequence of cells
  storing the elements.

  The idea of combining iterators and const iterators into the same
  class is taken from [Matt Austern. Defining iterators and const
  iterators. C/C++ User's Journal 19,1 (2001), 74-79].

  Author: Jyrki Katajainen, Bo Simonsen, April 2008
*/

#ifndef __CPHSTL_ENTRY_ITERATOR__
#define __CPHSTL_ENTRY_ITERATOR__

#include <iterator> // defines std::random_access_iterator_tag
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <utility> // defines std::pair

#include "type.h++" // defines cphstl::if_then_else

namespace cphstl {

  template <typename V, typename A, typename R, typename I, typename J>
  class vector;
  template <typename V, typename A, typename R, typename I, typename J>
  class ps_vector;
  template <typename I, typename A, typename E>
  class snapshot_iterator;


  template <typename R, typename E, bool is_const = false>
  class entry_iterator {

  public:

    // types

    typedef std::random_access_iterator_tag iterator_category;
    typedef typename R::value_type value_type;
    typedef std::size_t size_type;
    typedef std::ptrdiff_t difference_type;
    typedef typename if_then_else<is_const, value_type const*, value_type*>::type pointer;
    typedef typename if_then_else<is_const, value_type const&, value_type&>::type reference;

    typedef E entry;

  protected:

    // types

    typedef typename if_then_else<is_const, E const*, E*>::type node_pointer;

  public:

    // friends

    friend class entry_iterator<R, E, !is_const>;

    template <typename V, typename A, typename S, typename I, typename J>
    friend class cphstl::vector;
    template <typename V, typename A, typename S, typename I, typename J>
    friend class cphstl::ps_vector;


    template <typename I, typename A>
    friend class cphstl::snapshot_iterator;

    // structors

    entry_iterator();
    entry_iterator(entry_iterator<R, E, false> const&);
    entry_iterator(entry_iterator<R, E, true> const&);
    entry_iterator& operator=(entry_iterator const&);
    ~entry_iterator();

    // operators

    reference operator*() const;
    pointer operator->() const;
    entry_iterator& operator++();
    entry_iterator operator++(int);
    entry_iterator& operator--();
    entry_iterator operator--(int);
    entry_iterator& operator+=(difference_type);
    entry_iterator& operator-=(difference_type);
    entry_iterator operator+(difference_type) const;
    entry_iterator operator-(difference_type) const;
    difference_type operator-(entry_iterator const&) const;

    template <bool both>
    bool operator==(entry_iterator<R, E, both> const&) const;

    template <bool both>
    bool operator!=(entry_iterator<R, E, both> const&) const;

    template <bool both>
    bool operator<(entry_iterator<R, E, both> const&) const;

    template <bool both>
    bool operator>(entry_iterator<R, E, both> const&) const;

    template <bool both>
    bool operator<=(entry_iterator<R, E, both> const&) const;

    template <bool both>
    bool operator>=(entry_iterator<R, E, both> const&) const;

  protected:
    // converters to be used by the container friends
    entry_iterator(node_pointer const&);
    operator node_pointer() const;

    node_pointer position;  
  };

  template<typename R, typename E, bool both>
  entry_iterator<R, E, both> 
  operator+(typename R::difference_type, entry_iterator<R, E, both> const&);

}

#include "entry_iterator.c++" // implements cphstl::entry_iterator

#endif 
