/*! \file
 * \brief This file defines the function <code>is_sorted</code>.
 *
 * <h3>
 * Original author
 * </h3>
 *
 * Jyrki Katajainen <jyrki@diku.dk>, October - November 2000, November 2002
 * Bo Simonsen <bosim@diku.dk>, April 2008
 *
 */

#include <iterator> /* defines std::iterator_traits, std::input_iterator_tag, 
                       and std::random_access_iterator_tag */
#include <functional> // defines std::less
#include "assert.h++"

#ifdef CONCEPT
#include <concepts>
#endif

#ifdef CONCEPT
    concept InputIterator<typename Iter> {
      typename value_type;
      typename iterator_category;
      typename difference_type;
      typename pointer;
      typename reference;
      Iter& operator++(Iter& x);
      value_type& operator*(Iter);
      bool operator==(Iter, Iter&);
      bool operator!=(Iter, Iter&);
    };
    concept RandomAccessIterator<typename Iter> {
      typename value_type;
      typename iterator_category;
      typename difference_type;
      typename pointer;
      typename reference;
      Iter& operator++(Iter& x);
      value_type& operator*(Iter);
      bool operator==(Iter, Iter);
      bool operator<(Iter, Iter);
      Iter operator-(Iter, difference_type);
      difference_type operator-(Iter, Iter&);
      bool operator!=(Iter, Iter);
    };
    concept Container<typename Cont> {
      typename value_type;
      typename reference;
      typename pointer;
      typename iterator;
      iterator begin();
      iterator end();
    };
    concept Comparator<typename T> {
      typename first_argument_type;
      typename second_argument_type;
      bool operator()(T&, first_argument_type x, second_argument_type& y);
    };

    concept Assignable<typename T> {
      T& operator=(T& x, T y);
    };

#endif

namespace cphstl {
 
  namespace helper {

    /*! \brief Function <code>is_sorted</code> for input iterators.
     */
    // Requirements
    // iterator should be of type input_iterator
    // input_iterator should offer operator++ (prefix so no (int) argument), operator==, operator!= and operator*
    // comparator should offer bool operator()(const T& x, const T& y) 
    // element (std::iterator_traits<input_iterator>::value_type) should offer operator<, operator=

#ifdef CONCEPT
    template <InputIterator input_iterator, Comparator comparator>
    requires std::CopyConstructible<input_iterator::value_type>, Assignable<input_iterator::value_type>
#else
    template <typename input_iterator, typename comparator>
#endif 

    inline 
    bool 
    is_sorted (
      input_iterator p, 
      input_iterator one_past_the_end, 
      comparator less,
      std::input_iterator_tag // used for branching
    ) {
#ifndef CONCEPT
      typedef typename std::iterator_traits<input_iterator>::value_type element;    
#else
      typedef input_iterator::value_type element;
#endif

#ifdef SA    
      /* Will fail if less doesn't return bool */
      static_assert(sizeof(less(*p, *one_past_the_end)) == sizeof(bool), less_dont_return_bool);
      /* Check operators */
      static_assert(sizeof(++p) == sizeof(input_iterator), iterator_dont_return_same_type);
      static_assert(sizeof(*p) == sizeof(element), iterator_dont_return_element_type);
      static_assert(sizeof(p) == sizeof(one_past_the_end), iterators_not_of_same_type);
      static_assert(sizeof(p == one_past_the_end) == sizeof(bool), equal_operator_dont_return_bool)
#endif 

      if (p == one_past_the_end) { 
         return true;
      }
      
      element e = *p;
      ++p;
      if (p == one_past_the_end) {
        return true;
      }
      do {
        element f = *p;
        if (less(f, e)) {
           return false;
        }
        ++p;
        if (p == one_past_the_end) {
          break;
        }
        e = *p;
        if (less(e, f)) {
          return false;
        }
        ++p;
      } while (p != one_past_the_end);
      /* Bo: Was p == one_past_the_end, but doesn't make sense, my explaination
         of why nobody found this error was that they only used the code below,
         with the random access iterator */
      return true;
    }

    /*! \brief Function <code>is_sorted</code> for random access iterators.
     */

    // Requirements
    // iterator should be random_access_iterator
    // iterator should have operator>=, operator++ (prefix, no (int)), operator!=, operator-, operator*
    // element should offer operator= and < (comparator)
    // 

#ifdef CONCEPT
    template <RandomAccessIterator random_access_iterator, Comparator comparator>
    requires std::CopyConstructible<random_access_iterator::value_type>, Assignable<random_access_iterator::value_type>
#else
    template <typename random_access_iterator, typename comparator>
#endif
    inline 
    bool 
    is_sorted (
      random_access_iterator p, 
      random_access_iterator one_past_the_end, 
      comparator less,
      std::random_access_iterator_tag // used for branching
    ) {
#ifndef CONCEPT
      typedef typename std::iterator_traits<random_access_iterator>::value_type element;
      typedef typename std::iterator_traits<random_access_iterator>::difference_type index;
#else
      typedef random_access_iterator::value_type element;
      typedef random_access_iterator::difference_type index;
#endif

#ifdef SA
      /* Will fail if less doesn't return bool */
      static_assert(sizeof(less(*p, *one_past_the_end)) == sizeof(bool), less_dont_return_bool);
      /* Check operators */
      static_assert(sizeof(++p) == sizeof(random_access_iterator), iterator_dont_return_same_type);
      static_assert(sizeof(*p) == sizeof(element), iterator_dont_return_element_type);
      static_assert(sizeof(p) == sizeof(one_past_the_end), iterators_not_of_same_type);
      static_assert(sizeof(p == one_past_the_end) == sizeof(bool), equal_operator_dont_return_bool)
#endif

      if (p == one_past_the_end) { 
	return true;
      }
      
      dynamic_assert((p < one_past_the_end));
      ++p;
      
      index pairs_left = (one_past_the_end - p) / 2;
      
      element e = *(p - 1);
      
      while (pairs_left > 0) {
        element f = *p;
        if (less(f, e)) {
          return false;
        }
        ++p;
        e = *p;
        if (less(e, f)) {
          return false;
        }
        ++p;
        pairs_left -= 1;
      }
      
      if (p != one_past_the_end) {
        return !less(*p, e);
      }
      
      return true;
    }
  }  

  /*! \ingroup algorithm
   * \brief Function <code>is_sorted</code>
   *
   * <h3>
   * Input
   * </h3>
   *
   * A sequence represented by two iterators and an ordering function.
   *
   * <h3>
   * Effect
   * </h3>
   *
   * Checks whether the given sequence is in sorted order with respect to the
   * given ordering function.
   * 
   * <h3>
   * Requirements
   * </h3>
   *
   * <ul>
   * <li> 
   * The type <code>iterator</code> must have the capabilities of an input iterator.
   * </li>
   *
   * <li> 
   * It must be possible to compare the equality of any object referred to by an
   * <code>iterator</code> and an <code>element</code>.
   * </li>
   * </ul>
   *
   * <h3>
   * Preconditions
   * </h3>
   *
   * <ul>
   * <li> 
   * [<code>p</code>, <code>one_past_the_end</code>) must be a valid range.
   * </li>
   * </ul>
   *
   * <h3>
   * Implementation
   * </h3>
   *
   * The function appears in two versions depending on the iterator category of
   * the type <code>iterator</code>. 
   *
   * <h3>
   * Performance
   * </h3>
   *
   * <ul>
   * <li>
   * Exactly <code>one_past_the_end - p - 1</code> element comparisons are 
   * performed.
   * </li>
   *
   * <li> 
   * Since the sequence is scanned sequentially, the cache performance is 
   * optimal if the elements are stored sequentially.
   * </li>
   *
   * </ul>
   *
   * <h3>
   * Known Bugs
   * </h3>
   *
   * This version still relies on std::.
   *
   */

  /* Iterator overload */
  // Requirements
  // I should have the types value_type, iterator_category.
  // I::value_type should have operator<.  
  // (Reference: http://www.ib.cnea.gov.ar/~oop/biblio/libstdc++/structstd_1_1iterator__traits.html)

  template <typename I>
  inline bool 
  is_sorted(I first, I one_past_the_end) {
    typedef typename std::iterator_traits<I>::value_type element;
    typedef typename std::iterator_traits<I>::iterator_category category;
    return cphstl::helper::is_sorted(first, one_past_the_end, std::less<element>(), category());
  }

  // Requirements
  // Same as above method AND
  // O requires bool operator()(const T& x, const T& y) 
  
  template <typename I, typename O>
  inline bool 
  is_sorted ( 
    I first, 
    I one_past_the_end, 
    O less
  ) {
    typedef typename std::iterator_traits<I>::value_type element;
    typedef typename std::iterator_traits<I>::iterator_category category;
    return cphstl::helper::is_sorted(first, one_past_the_end, less, category());
  }

  /* Container overload */
  
  // Requirements:
  // C has begin(), which returns a iterator.
  // C has end(), which returns a iterator.
  // C::iterator should have the types value_type, iterator_category.
  // C::iterator<iterator>::value_type should have operator<.
  // (Reference: http://www.ib.cnea.gov.ar/~oop/biblio/libstdc++/structstd_1_1iterator__traits.html)
 
#ifdef CONCEPT
  template <Container C>
#else
  template <typename C>
#endif
  inline bool 
  is_sorted(C cont) {

#ifndef CONCEPT
    typedef typename std::iterator_traits<typename C::iterator>::value_type element;
    typedef typename std::iterator_traits<typename C::iterator>::iterator_category category;
#else
    typedef C::iterator::value_type element;
    typedef C::iterator::iterator_category category;
#endif

#ifdef SA
    static_assert(sizeof(cont.begin()) == sizeof(typename C::iterator), begin_doesnt_return_iterator);
    static_assert(sizeof(cont.end()) == sizeof(typename C::iterator), end_doesnt_return_iterator);    
#endif
    
    return cphstl::helper::is_sorted(cont.begin(), cont.end(), std::less<element>(), category());
  }

  // Requirements
  // Same as above method AND
  // O requires bool operator()(const T& x, const T& y) 
  
#ifdef CONCEPT
  template <Container C, Comparator O>
#else
  template <typename C, typename O>
#endif
  inline bool 
  is_sorted ( 
    C cont,
    O less
  ) {
#ifndef CONCEPT
    typedef typename std::iterator_traits<typename C::iterator>::value_type element;
    typedef typename std::iterator_traits<typename C::iterator>::iterator_category category;
#else
    typedef C::iterator::value_type element;
    typedef C::iterator::iterator_category category;
#endif

#ifdef SA
    static_assert(sizeof(cont.begin()) == sizeof(typename C::iterator), begin_doesnt_return_iterator);
    static_assert(sizeof(cont.end()) == sizeof(typename C::iterator), end_doesnt_return_iterator);    
#endif    

    return cphstl::helper::is_sorted(cont.begin(), cont.end(), less, category());
  }

  

}
