/*
  A list is a bridge that delegates the work to the given realization.

  Author: Jyrki Katajainen, April 2008
*/

#include <cassert> // defines assert
#include <functional> // defines std::bind2nd, std::equal_to, and std::less
#include <iostream> // defines std::cout
#include "type.h++" // defines cphstl::type

namespace cphstl {

  // default constructor

  template <typename V, typename A, typename R>
  list<V, A, R>::list(A const& a) 
    : kernel(a) {
 }

  // parametrized constructors

  template <typename V, typename A, typename R>
  list<V, A, R>::list(size_type s, V const& v, A const& a) 
    : kernel(a) { 
    (*this).insert((*this).end(), size_type(s), v);
  }

  template <typename V, typename A, typename R>
  template <typename I>
  list<V, A, R>::list(I p, I q, A const& a)
    : kernel(a) {
    (*this).insert((*this).end(), p, q);
  }

  // copy constructor

  template <typename V, typename A, typename R>
  list<V, A, R>::list(list const& other) : kernel(other.kernel) { 
  }

  // destructor

  template <typename V, typename A, typename R>
  list<V, A, R>::~list() { 
  }

  // operator=

  template <typename V, typename A, typename R>
  list<V, A, R>& 
  list<V, A, R>::operator=(list const& other) {
    (*this).kernel = other.kernel;
    return *this;
  }

  // assign; function overloading can be a pain!

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::assign(size_type n, V const& v) {
    (*this).assign_fill(n, v);
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::assign(I p, I q) {
    (*this).assign_dispatch(p, q, cphstl::int2type<cphstl::type<I>::is_integral>());
  }

  template <typename V, typename A, typename R>
  template <typename Integer>
  void
  list<V, A, R>::assign_dispatch(Integer n, Integer v, cphstl::int2type<1>) {
    (*this).assign_fill(static_cast<size_type>(n), static_cast<V>(v));
  }

  template <typename V, typename A, typename R>
  template <typename Iterator>
  void
  list<V, A, R>::assign_dispatch(Iterator p, Iterator q, cphstl::int2type<0>) {
    (*this).assign_range(p, q);
  }

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::assign_fill(size_type n, V const& v) {
    A a = (*this).get_allocator();
    R temporary = R(a);
    temporary.insert(temporary.begin(), n, v);
    (*this).erase((*this).begin(), (*this).end());
    (*this).kernel.swap(temporary);
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::assign_range(I p, I q) {
    A a = (*this).get_allocator();
    R temporary = R(a);
    temporary.insert(temporary.begin(), p, q);
    (*this).erase((*this).begin(), (*this).end());
    (*this).kernel.swap(temporary);
  }

  // begin

  template <typename V, typename A, typename R>
  typename list<V, A, R>::iterator 
  list<V, A, R>::begin() {
    return iterator((*this).kernel.begin());
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_iterator 
  list<V, A, R>::begin() const {
    return const_iterator((*this).kernel.begin());
  }

  // end

  template <typename V, typename A, typename R>
  typename list<V, A, R>::iterator
  list<V, A, R>::end() {
    return iterator((*this).kernel.end());
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_iterator 
  list<V, A, R>::end() const {
    return const_iterator((*this).kernel.end());
  }

  // rbegin

  template <typename V, typename A, typename R>
  typename list<V, A, R>::reverse_iterator 
  list<V, A, R>::rbegin() {
    return reverse_iterator((*this).end());
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_reverse_iterator 
  list<V, A, R>::rbegin() const {
    return const_reverse_iterator((*this).end());
  }

  // rend

  template <typename V, typename A, typename R>
  typename list<V, A, R>::reverse_iterator 
  list<V, A, R>::rend() {
    return reverse_iterator((*this).begin());
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_reverse_iterator 
  list<V, A, R>::rend() const {
    return const_reverse_iterator((*this).begin());
  }

  // get_allocator

  template <typename V, typename A, typename R>
  A
  list<V, A, R>::get_allocator() const {
    return (*this).kernel.get_allocator();
  }

  // size

  template <typename V, typename A, typename R>
  typename list<V, A, R>::size_type 
  list<V, A, R>::size() const {
    return (*this).kernel.size();
  }

  // max_size

  template <typename V, typename A, typename R>
  typename list<V, A, R>::size_type 
  list<V, A, R>::max_size() const {
    return (*this).kernel.max_size();
  }

  // empty

  template <typename V, typename A, typename R>
  bool 
  list<V, A, R>::empty() const {
    return (*this).size() == size_type(0);
  }

  // front

  template <typename V, typename A, typename R>
  typename list<V, A, R>::reference
  list<V, A, R>::front() {
    iterator const first = (*this).begin();
    return *first;
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_reference
  list<V, A, R>::front() const {
    const_iterator const first = (*this).begin();
    return *first;
  }

  // back

  template <typename V, typename A, typename R>
  typename list<V, A, R>::reference
  list<V, A, R>::back() {
    iterator const last = --(*this).end();
    return *last;
  }

  template <typename V, typename A, typename R>
  typename list<V, A, R>::const_reference
  list<V, A, R>::back() const {
    const_iterator const last = --(*this).end();
    return *last;
  }

  // resize

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::resize(size_type s, V v) {
    if (s > (*this).size()) {
      (*this).kernel.insert((*this).kernel.end(), s - (*this).size(), v);
    }
    else {
      while (s < (*this).size()) {
	(*this).pop_back();
      }
    }
  }

  // push_front

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::push_front(V const& v) {
    (*this).insert((*this).begin(), v);
  }

  // pop_front

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::pop_front() {
    (*this).erase((*this).begin());
  }

  // push_back

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::push_back(V const& v) {
    (*this).insert((*this).end(), v);
  }

  // pop_back

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::pop_back() {
    (*this).erase(--(*this).end());
  }

  // single-element insert

  template <typename V, typename A, typename R>
  typename list<V, A, R>::iterator
  list<V, A, R>::insert(iterator p, V const& v) {
    return iterator((*this).kernel.insert(p, v));
  }

  // multiple-element inserts

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::insert(iterator p, size_type n, V const& v) {
    (*this).insert_fill(p, n, v);
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::insert(iterator p, I q, I r) {
    (*this).insert_dispatch(p, q, r, cphstl::int2type<cphstl::type<I>::is_integral>());
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::insert_dispatch(iterator p, I n, I v, cphstl::int2type<1>) {
    (*this).insert_fill(p, static_cast<size_type>(n), static_cast<V>(v));
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::insert_dispatch(iterator p, I q, I r, cphstl::int2type<0>) {
    (*this).insert_range(p, q, r);
  }

  template <typename V, typename A, typename R>
  void
  list<V, A, R>::insert_fill(iterator p, size_type s, V const& v) {
    (*this).kernel.insert(p, s, v);
  }

  template <typename V, typename A, typename R>
  template <typename I>
  void
  list<V, A, R>::insert_range(iterator p, I q, I r) {
    (*this).kernel.insert(p, q, r);
  }

  // single-element erase

  template <typename V, typename A, typename R>
  typename list<V, A, R>::iterator
  list<V, A, R>::erase(iterator p) {
    return iterator((*this).kernel.erase(p));
  }

  // multiple-element erase

  template <typename V, typename A, typename R>
  typename list<V, A, R>::iterator
  list<V, A, R>::erase(iterator p, iterator q) {
    while (p != q) {
      p = iterator((*this).erase(p));
    }
    return q;
  }

  // clear

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::clear() {
    (void)(*this).erase((*this).begin(), (*this).end());
  }

  // swap

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::swap(list<V, A, R>& r) {
    (*this).kernel.swap(r.kernel);
  }

  template <typename V, typename A, typename R>
  void 
  swap(list<V, A, R>& r, list<V, A, R>& s) {
    r.swap(s);
  }

  // splice

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::splice(iterator p, list<V, A, R>& x) {
    assert(&x != this);
    (*this).kernel.splice(p, x.kernel);
  }

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::splice(iterator p, list<V, A, R>& x, iterator q) {
    assert(q != x.end());
    (*this).kernel.splice(p, x.kernel, q);
  }

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::splice(iterator p, list<V, A, R>& x, iterator q, iterator r) {
    (*this).kernel.splice(p, x.kernel, q, r);
  }

  // remove

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::remove(V const& v) {
    typedef std::equal_to<V> E;
    E b = E();
    std::binder2nd<E> u = std::binder2nd<E>(b, v);
    (*this).remove_if(u);
  }

  // remove_if

  template <typename V, typename A, typename R>
  template <typename P>
  void 
  list<V, A, R>::remove_if(P unary_predicate) {
    (*this).kernel.remove_if(unary_predicate);
  }

  // unique

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::unique() {
    (*this).unique(std::equal_to<V>());
  }

  template <typename V, typename A, typename R>
  template <typename P>
  void 
  list<V, A, R>::unique(P binary_predicate) {
    (*this).kernel.unique(binary_predicate);
  }

  // merge

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::merge(list<V, A, R>& x) {
    (*this).merge(x, std::less<V>());
  }

  template <typename V, typename A, typename R>
  template <typename C>
  void 
  list<V, A, R>::merge(list<V, A, R>& x, C c) {
    (*this).kernel.merge(x.kernel, c);
  }

  // sort

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::sort() {
    (*this).sort(std::less<V>());
  }

  template <typename V, typename A, typename R>
  template <typename C>
  void 
  list<V, A, R>::sort(C c) {
    (*this).kernel.sort(c);
  }

  // reverse

  template <typename V, typename A, typename R>
  void 
  list<V, A, R>::reverse() {
    (*this).kernel.reverse();
  }

  // operator==

  template <typename V, typename A, typename R>
  bool
  operator==(list<V, A, R> const& r, list<V, A, R> const& s) {
    return (r.size() == s.size() && std::equal(r.begin(), r.end(), s.begin()));
  }

  // operator<

  template <typename V, typename A, typename R>
  bool
  operator<(list<V, A, R> const& r, list<V, A, R> const& s) {
    return std::lexicographical_compare(r.begin(), r.end(), s.begin(), s.end());
  }

  // operator!=

  template <typename V, typename A, typename R>
  bool
  operator!=(list<V, A, R> const& r, list<V, A, R> const& s) {
    return !(r == s);
  }

  // operator>

  template <typename V, typename A, typename R>
  bool
  operator>(list<V, A, R> const& r, list<V, A, R> const& s) {
    return (s < r);
  }

  // operator>=

  template <typename V, typename A, typename R>
  bool
  operator>=(list<V, A, R> const& r, list<V, A, R> const& s) {
    return !(r < s);
  }

  // operator<=

  template <typename V, typename A, typename R>
  bool
  operator<=(list<V, A, R> const& r, list<V, A, R> const& s) {
    return !(s < r);
  }

}
