/*
  This header is taken quite directly from the C++ standard [2003].
  According to the standard, a list is a sequence that supports
  bidirectional iterators. More precisely, a list satisfies all of the
  requirements of a container, of a reversible container, and of a
  sequence, including most of the optional sequence requirements,
  exceptions being the operator[] and at member functions.

  Author: Jyrki Katajainen, April 2008
*/

#ifndef __CPHSTL_LIST__
#define __CPHSTL_LIST__

#include <iterator> // defines std::reverse_iterator
#include <memory> // defines std::allocator
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <functional> // defines std::less
#include <algorithm> // defines std::equal and std::lexicographical_compare
#include <list> // defines std::list

#include "type.h++" // defines cphstl::int2type
#include "doubly_linked_list.h++" // defines cphstl::doubly_linked_list

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>,
    typename R = std::list<V, A> 
  >
  class list {

  public:

    // types
    
    typedef V value_type;
    typedef A allocator_type;
    typedef V& reference;
    typedef V const& const_reference;
    typedef V* pointer;
    typedef V const* const_pointer;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;
    typedef typename R::iterator iterator;
    typedef typename R::const_iterator const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

    // structors
  
    explicit list(A const& = A());
    explicit list(size_type, V const& = V(), A const& = A());
  
    template <typename I>
    list(I, I, A const& = A());
  
    list(list const&);
    ~list();
    list& operator=(list const&);

    void assign(size_type, V const&);

    template <typename I>
    void assign(I, I);

    // iterators

    iterator begin();
    const_iterator begin() const;
    iterator end();
    const_iterator end() const;
    reverse_iterator rbegin();
    const_reverse_iterator rbegin() const;
    reverse_iterator rend();
    const_reverse_iterator rend() const;

    // accessors

    A get_allocator() const;
    size_type size() const;
    size_type max_size() const;
    bool empty() const;
    V const& front() const;
    V const& back() const;

    // modifiers

    V& front();
    V& back();
    void resize(size_type, V = V());
    void push_front(V const&);
    void pop_front();
    void push_back(V const&);
    void pop_back();
    iterator insert(iterator, V const&);
    void insert(iterator, size_type, V const&);

    template <typename I>
    void insert(iterator, I, I);

    iterator erase(iterator);
    iterator erase(iterator, iterator);
    void clear();
    void swap(list&);

    // list operations

    void splice(iterator, list&);
    void splice(iterator, list&, iterator);
    void splice(iterator, list&, iterator, iterator);
    void remove(V const&);

    template <typename P> 
    void remove_if(P);

    void unique();

    template <typename P>
    void unique(P);

    void merge(list&);

    template <typename C> 
    void merge(list&, C);

    void sort();

    template <typename C> 
    void sort(C);

    void reverse();
  
  protected:

    template <typename I>
    void assign_dispatch(I, I, cphstl::int2type<0>);

    template <typename I>
    void assign_dispatch(I, I, cphstl::int2type<1>);

    void assign_fill(size_type, V const&);

    template <typename I>
    void assign_range(I, I);

    template <typename I>
    void insert_dispatch(iterator, I, I, cphstl::int2type<0>);

    template <typename I>
    void insert_dispatch(iterator, I, I, cphstl::int2type<1>);

    void insert_fill(iterator, size_type, V const&);

    template <typename I>
    void insert_range(iterator, I, I);

    typedef R realization_type;
    R kernel; 

  };

  template <typename V, typename A, typename R>
  void swap(list<V, A, R>&, list<V, A, R>&);
  
  template <typename V, typename A, typename R>
  bool operator==(list<V, A, R> const&, list<V, A, R> const&);
  
  template <typename V, typename A, typename R>
  bool operator<(list<V, A, R> const&, list<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator!=(list<V, A, R> const&, list<V, A, R> const&);
  
  template <typename V, typename A, typename R>
  bool operator>(list<V, A, R> const&, list<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator>=(list<V, A, R> const&, list<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator<=(list<V, A, R> const&, list<V, A, R> const&);

}

#include "list.c++" // implements cphstl::list

#endif
