/*
  cphstl::list_based_stack is a simple realization of a stack. This
  implementation gives the following guarantees:
  - performance bounds specified in  the C++ standard [2003] are satisfied,
  - iterators are kept valid at all times, and
  - all operations are strongly exception safe.

  Authors: Jyrki Katajainen and Johannes Serup, March 2008
*/

#ifndef __CPHSTL_LIST_BASED_STACK__
#define __CPHSTL_LIST_BASED_STACK__

#include <memory> // defines std::allocator
#include <list> // defines std::list
#include <cstddef> // defines std::size_t and std::ptrdiff_t

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>, 
    typename R = std::list<V, A>,
    typename I = typename R::iterator,
    typename J = typename R::const_iterator
  >
  class list_based_stack {
  public:

    // types

    typedef V value_type;
    typedef A allocator_type;
    typedef std::size_t size_type;
    typedef I iterator;
    typedef J const_iterator;

    // structors

    explicit list_based_stack(A const& = A()); 
    list_based_stack(list_based_stack<V, A, R, I, J> const&);
    ~list_based_stack();
    list_based_stack<V, A, R, I, J>& operator=(list_based_stack<V, A, R, I, J> const&); 

    // iterators

    iterator begin();
    const_iterator begin() const;
    iterator end();
    const_iterator end() const;

    // accessors

    size_type size() const;
    size_type max_size() const;
    V const& top() const;

    // modifiers

    V& top();
    void push(V const&);
    void pop();
    void clear();
    void swap(list_based_stack<V, A, R, I, J>&);
    
  protected:

    typedef R realization_type;
    realization_type list;
  };
  
} // namespace cphstl

#include "list_based_stack.c++" // implements cphstl::list_based_stack

#endif // __CPHSTL_LIST_BASED_STACK__
