/*
  The idea of combining iterators and const iterators into the same
  class is taken from [Matt Austern. Defining iterators and const
  iterators. C/C++ User's Journal 19,1 (2001), 74-79].

  Author: Jyrki Katajainen, Bo Simonsen, May 2006, April 2008

  General design idea is proposed by J. Katajainen and B. Simonsen
*/

#ifndef __CPHSTL_NODE_ITERATOR__
#define __CPHSTL_NODE_ITERATOR__

#include <iterator> // defines std::bidirectional_iterator_tag
#include <cstddef> // defines std::ptrdiff_t
#include <iostream> // defines std::ostream
#include <string> // defines std::string

#include "type.h++" // defines cphstl::if_then_else

namespace cphstl {
 
  /* Forward declarations of bridge classes */

  template <typename V, typename A, typename R>
  class list;

  template <typename V, typename C, typename A, typename R>
  class meldable_priority_queue;

  template <typename V, typename C, typename A, typename R, typename I, typename J>
  class multiset;

  template <typename V, typename C, typename A, typename R, typename I, typename J>
  class set;
  
  template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
  class set_base;

  template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
  class map;

  template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
  class map_base;


  template <typename N, bool is_const = false>
  class node_iterator {

  public:

    // types

    typedef std::bidirectional_iterator_tag iterator_category;
    typedef typename N::value_type value_type;
    typedef std::ptrdiff_t difference_type;

    typedef typename if_then_else<is_const, value_type const*, value_type*>::type pointer;
    typedef typename if_then_else<is_const, value_type const&, value_type&>::type reference;

    typedef node_iterator<N, !is_const> complement;
  private:

    // types

    typedef typename if_then_else<is_const, N const*, N*>::type node_pointer;

  public:

    // friends
    friend class node_iterator<N, !is_const>;
    
    template <typename M, bool both>
    friend std::ostream& operator<<(std::ostream&, node_iterator<M, both> const&);

    template <typename V, typename A, typename R>
    friend class cphstl::list;
    
    template <typename V, typename C, typename A, typename R>
    friend class cphstl::meldable_priority_queue;
    
    template <typename V, typename C, typename A, typename R, typename I, typename J>
    friend class cphstl::set;
    
    template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
    friend class cphstl::set_base;
 
    template <typename V, typename C, typename A, typename R, typename I, typename J>
    friend class cphstl::multiset;
 
    template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
    friend class cphstl::map;

    template <typename K, typename V, typename C, typename A, typename R, typename I, typename J>
    friend class cphstl::map_base;
  
    template <typename T1, typename T2> 
    friend class std::pair;

    // structors

    node_iterator();
    node_iterator(node_iterator<N, false> const&);
    node_iterator& operator=(node_iterator const&);
    ~node_iterator();

    // operators

    reference operator*() const;
    pointer operator->() const;
    node_iterator& operator++();
    node_iterator operator++(int);
    node_iterator& operator--();
    node_iterator operator--(int);

    template <bool both>
    bool operator==(node_iterator<N, both> const&) const;

    template <bool both>
    bool operator!=(node_iterator<N, both> const&) const;

  private:
    // converters to be used by the friends

    node_iterator(node_pointer); // node_pointer --> iterator 
    operator node_pointer() const; // iterator --> node_pointer
    operator std::string() const; // iterator --> string

    node_pointer link;
    
  };

}

#include "node_iterator.c++" // implements cphstl::node_iterator

#endif
