/*
  According to the C++ standard [2003], a stack is an adaptor for a
  sequence supporting functions back, push_back, and pop_back.  A CPH
  STL extension of a stack is a reversible container that supports
  bidirectional iterators. That is, cphstl:.stack satisfies all the
  requirements of a container and of a reversible container.  The
  iterators access the elements starting from top of the stack.

  Authors: Jyrki Katajainen and Johannes Serup, March 2008
*/

#ifndef __CPHSTL_STACK__
#define __CPHSTL_STACK__

#include <memory> // defines std::allocator
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <iterator> // defines std::reverse_iterator
#include "list_based_stack.h++" // defines cphstl::list_based_stack

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>, 
    typename R = cphstl::list_based_stack<V, A> 
  >
  class stack {
  public:

    // types

    typedef V value_type;
    typedef A allocator_type;
    typedef V& reference;
    typedef V const& const_reference;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;
    typedef typename R::iterator iterator;
    typedef typename R::const_iterator const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
 
    // structors

    explicit stack(A const& = A()); 
    stack(stack<V, A, R> const&);
    ~stack();
    stack<V, A, R>& operator=(stack<V, A, R> const&); 

    // iterators

    iterator begin();
    const_iterator begin() const;
    iterator end();
    const_iterator end() const;
    reverse_iterator rbegin();
    const_reverse_iterator rbegin() const;
    reverse_iterator rend();
    const_reverse_iterator rend() const;

    // accessors

    size_type size() const;
    size_type max_size() const;
    bool empty() const;
    V const& top() const;

    // modifiers

    V& top();
    void push(V const&);
    void pop();
    void clear();
    void swap(stack<V, A, R>&);

  protected:

    typedef R realization_type;
    realization_type kernel;    

  };

  template <typename V, typename A, typename R>
  void swap(stack<V, A, R>&, stack<V, A, R>&);

  template <typename V, typename A, typename R>
  bool operator==(stack<V, A, R> const&, stack<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator<(stack<V, A, R> const&, stack<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator!=(stack<V, A, R> const&, stack<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator>(stack<V, A, R> const&, stack<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator>=(stack<V, A, R> const&, stack<V, A, R> const&);

  template <typename V, typename A, typename R>
  bool operator<=(stack<V, A, R> const&, stack<V, A, R> const&);

}

#include "stack.c++" // implements cphstl::stack

#endif
