/*
  Implementation of cphstl::persistent_set

  The persistent set class is an extention of the set bridge class.
  With partial persistent data structures, can we get iterators for
  any revision, at any time, which is what this bridge class supports
  using the persistent_set_adapter which is instatiated when the [] operator
  is used, such that we can do this:

    persistent_set<....> s;
    ...
    // revision 2 and 5 start iterator similar for end.
    it = s[2].begin();
    it = s[5].begin();
    // current / lastest revision 
    it = s.begin();

  Author: Bo Simonsen, November 2008
*/



#include "stl_set.h++"

namespace cphstl {

  template <typename R, typename I, typename J>
  class persistent_set_adapter {
  public:
    typedef std::size_t size_type;
    typedef R realization_type;
    typedef I iterator;
    typedef J const_iterator;
  private:
    realization_type* link;
    size_type revision;
  public:
    /* pointer to avoid copy construction */
    explicit persistent_set_adapter(realization_type* r, size_type rev) {
      link = r;
      revision = rev;
    }
    size_type size() {
      return (*link).size(revision);
    }
    iterator begin() {
      return (*link).begin(revision);
    }
    iterator end() {
      return (*link).end(revision);
    }
    const_iterator begin() const {
      return (*link).begin(revision);
    }
    const_iterator end() const {
      return (*link).end(revision);
    }

  };

  template <
    typename V, 
    typename C = std::less<V>,
    typename A = std::allocator<V>,
    typename R = std::set<V, C, A>,
    typename I = typename R::iterator, 
    typename J = typename R::const_iterator
  >
  class persistent_set : public set<V, C, A, R, I, J> {
  public:
    typedef set<V, C, A, R, I, J> superclass;
    typedef std::size_t size_type;
    typedef I iterator;
    typedef J const_iterator;

    size_type size() {
      return (*this).kernel.size((*this).kernel.revision());
    }

    iterator begin() {
      return (*this).kernel.begin((*this).kernel.revision());
    }
    iterator end() {
      return (*this).kernel.end((*this).kernel.revision());
    }
    const_iterator begin() const {
      return (*this).kernel.begin((*this).kernel.revision());
    }
    const_iterator end() const {
      return (*this).kernel.end((*this).kernel.revision());
    }
    persistent_set_adapter<R, I, J> operator[](size_type revision) {
      return persistent_set_adapter<R, I, J>(&((*this).kernel), revision);
    }

  };
}


