/*
  A vector is a bridge that delegates the work to the given
  realization.

  Authors: Tina A. G. Andersen, Filip Bruman, Jyrki Katajainen, Daniel
  P. Larsen, Claus Ullerlund, Christian Wolfgang, March 2008
*/

namespace cphstl {

  // explicit constructor

  template <typename V, typename A, typename R, typename I, typename J>
  vector<V, A, R, I, J>::vector(A const& a) 
    : kernel(a) {
 }

  // parametrized constructors

  template <typename V, typename A, typename R, typename I, typename J>
  vector<V, A, R, I, J>::vector(size_type s, V const& v, A const& a) 
    : kernel(a) { 
    (*this).insert((*this).end(), size_type(s), v);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  vector<V, A, R, I, J>::vector(K p, K q, A const& a)
    : kernel(a) {
    (*this).insert((*this).end(), p, q);
  }

  // copy constructor

  template <typename V, typename A, typename R, typename I, typename J>
  vector<V, A, R, I, J>::vector(vector const& other) : kernel(other.kernel) { 
  }

  // destructor

  template <typename V, typename A, typename R, typename I, typename J>
  vector<V, A, R, I, J>::~vector() { 
  }

  // operator=

  template <typename V, typename A, typename R, typename I, typename J>
  vector<V, A, R, I, J>& 
  vector<V, A, R, I, J>::operator=(vector const& other) {
    (*this).kernel = other.kernel;
    return *this;
  }

  // assign

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::assign(size_type n, V const& v) {
    (*this).assign_fill(n, v);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::assign(K p, K q) {
    (*this).assign_dispatch(p, q, cphstl::int2type<cphstl::type<K>::is_integral>());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename Integer>
  void
  vector<V, A, R, I, J>::assign_dispatch(Integer n, Integer v, cphstl::int2type<1>) {
    (*this).assign_fill(static_cast<size_type>(n), static_cast<V>(v));
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename Iterator>
  void
  vector<V, A, R, I, J>::assign_dispatch(Iterator p, Iterator q, cphstl::int2type<0>) {
    (*this).assign_range(p, q);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::assign_fill(size_type n, const V& v) {
    A a = (*this).kernel.get_allocator();
    R temporary = R(a);
    temporary.insert(temporary.begin(), size_type(n), v);
    (*this).kernel.erase((*this).begin(), (*this).end());
    (*this).kernel.swap(temporary);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::assign_range(K p, K q) {
    A a = (*this).kernel.get_allocator();
    R temporary = R(a);
    temporary.insert(temporary.begin(), p, q);
    (*this).kernel.erase((*this).begin(), (*this).end());
    (*this).kernel.swap(temporary);
  }

  // begin

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::iterator 
  vector<V, A, R, I, J>::begin() {
    return iterator((*this).kernel.begin());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_iterator 
  vector<V, A, R, I, J>::begin() const {
    return const_iterator((*this).kernel.begin());
  }

  // end

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::iterator
  vector<V, A, R, I, J>::end() {
    return (*this).begin() + difference_type((*this).kernel.size());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_iterator 
  vector<V, A, R, I, J>::end() const {
    return (*this).begin() + difference_type((*this).kernel.size());
  }

  // rbegin

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reverse_iterator 
  vector<V, A, R, I, J>::rbegin() {
    return reverse_iterator((*this).end());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reverse_iterator 
  vector<V, A, R, I, J>::rbegin() const {
    return const_reverse_iterator((*this).end());
  }

  // rend

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reverse_iterator 
  vector<V, A, R, I, J>::rend() {
    return reverse_iterator((*this).begin());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reverse_iterator 
  vector<V, A, R, I, J>::rend() const {
    return const_reverse_iterator((*this).begin());
  }

  // get_allocator

  template <typename V, typename A, typename R, typename I, typename J>
  A
  vector<V, A, R, I, J>::get_allocator() const {
    return (*this).kernel.get_allocator();
  }

  // size

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::size_type 
  vector<V, A, R, I, J>::size() const {
    return (*this).kernel.size();
  }

  // max_size

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::size_type 
  vector<V, A, R, I, J>::max_size() const {
    return (*this).kernel.max_size();
  }

  // capacity

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::size_type 
  vector<V, A, R, I, J>::capacity() const {
    return (*this).kernel.capacity();
  }

  // empty

  template <typename V, typename A, typename R, typename I, typename J>
  bool 
  vector<V, A, R, I, J>::empty() const {
    return (*this).size() == size_type(0);
  }

  // operator[]

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reference
  vector<V, A, R, I, J>::operator[](size_type s) {
    return (*this).kernel.operator[](s);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reference
  vector<V, A, R, I, J>::operator[](size_type s) const {
    return (*this).kernel.operator[](s);
  }

  // at

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reference
  vector<V, A, R, I, J>::at(size_type s) {
    if (s >= (*this).size()) {
      throw std::out_of_range("index out of bounds");
    }
    return (*this).operator[](s);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reference
  vector<V, A, R, I, J>::at(size_type s) const {
    if (s >= (*this).size()) {
      throw std::out_of_range("index out of bounds");
    }
    return (*this).operator[](s);
  }

  // front

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reference
  vector<V, A, R, I, J>::front() {
    iterator const first = (*this).kernel.begin();
    return *first;
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reference
  vector<V, A, R, I, J>::front() const {
    const_iterator const first = (*this).kernel.begin();
    return *first;
  }

  // back

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::reference
  vector<V, A, R, I, J>::back() {
    iterator const last = --(*this).end();
    return *last;
  }

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::const_reference
  vector<V, A, R, I, J>::back() const {
    const_iterator const last = --(*this).end();
    return *last;
  }

  // resize

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::resize(size_type s, V v) {
    if (s > (*this).size()) {
      (*this).kernel.insert((*this).end(), size_type(s - (*this).size()), v);
    }
    else if (s < (*this).size()) {
      (*this).kernel.erase((*this).begin() + difference_type(s), (*this).end());
    }
    else {
      // do nothing
    }
  }

  // reserve

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::reserve(size_type s) {
    (*this).kernel.reserve(s);
  }

  // push_back

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::push_back(V const& v) {
    (*this).kernel.insert((*this).end(), v);
  }

  // pop_back

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::pop_back() {
    iterator last = --(*this).end();
    (*this).kernel.erase(last);
  }

  // single-element insert

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::iterator
  vector<V, A, R, I, J>::insert(iterator p, V const& v) {
    return (*this).kernel.insert(p, v);
  }

  // multiple-element inserts

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::insert(iterator p, size_type n, V const& v) {
    (*this).insert_fill(p, n, v);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::insert(iterator p, K q, K r) {
    (*this).insert_dispatch(p, q, r, cphstl::int2type<cphstl::type<K>::is_integral>());
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::insert_dispatch(iterator p, K n, K v, cphstl::int2type<1>) {
    (*this).insert_fill(p, static_cast<size_type>(n), static_cast<V>(v));
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::insert_dispatch(iterator p, K q, K r, cphstl::int2type<0>) {
    (*this).insert_range(p, q, r);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  void
  vector<V, A, R, I, J>::insert_fill(iterator p, size_type s, V const& v) {
    (*this).kernel.insert(p, s, v);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  template <typename K>
  void
  vector<V, A, R, I, J>::insert_range(iterator p, K q, K r) {
    (*this).kernel.insert(p, q, r);
  }

  // single-element erase

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::iterator
  vector<V, A, R, I, J>::erase(iterator p) {
    return (*this).kernel.erase(p);
  }

  // multiple-element erase

  template <typename V, typename A, typename R, typename I, typename J>
  typename vector<V, A, R, I, J>::iterator
  vector<V, A, R, I, J>::erase(iterator p, iterator q) {
    return (*this).kernel.erase(p, q);
  }

  // clear

  template <typename V, typename A, typename R, typename I, typename J>
  void 
  vector<V, A, R, I, J>::clear() {
    (*this).kernel.erase((*this).begin(), (*this).end());
  }

  // swap

  template <typename V, typename A, typename R, typename I, typename J>
  void 
  vector<V, A, R, I, J>::swap(vector<V, A, R, I, J>& r) {
    (*this).kernel.swap(r.kernel);
  }

  template <typename V, typename A, typename R, typename I, typename J>
  void 
  swap(vector<V, A, R, I, J>& r, vector<V, A, R, I, J>& s) {
    r.swap(s);
  }

  // operator==

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator==(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return (r.size() == s.size() && std::equal(r.begin(), r.end(), s.begin()));
  }

  // operator<

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator<(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return std::lexicographical_compare(r.begin(), r.end(), s.begin(), s.end());
  }

  // operator!=

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator!=(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return !(r == s);
  }

  // operator>

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator>(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return (s < r);
  }

  // operator>=

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator>=(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return !(r < s);
  }

  // operator<=

  template <typename V, typename A, typename R, typename I, typename J>
  bool
  operator<=(vector<V, A, R, I, J> const& r, vector<V, A, R, I, J> const& s) {
    return !(s < r);
  }

}
