/*
  This header is taken quite directly from the C++ standard [2003].
  According to the standard, a vector is a sequence that supports
  random-access iterators. More precisely, a vector satisfies all of
  the requirements of a container, of a reversible container, and of a
  sequence, including most of the optional sequence requirements,
  exceptions being the push_front and pop_front functions.

  Authors: Tina A. G. Andersen, Filip Bruman, Jyrki Katajainen, Daniel
  P. Larsen, Claus Ullerlund, Christian Wolfgang, March 2008
*/

#ifndef __CPHSTL_VECTOR__
#define __CPHSTL_VECTOR__

#include <iterator> // defines std::reverse_iterator
#include <memory> // defines std::allocator
#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <algorithm> // defines std::equal and std::lexicographical_compare
#include <stdexcept> // defines std::length_error and std::out_of_range
#include <vector> // defines std::vector

#include "type.h++" // defines cphstl::int2type

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>,
    typename R = std::vector<V, A>,
    typename I = typename R::iterator,
    typename J = typename R::const_iterator
  >
  class vector {
  public:

    // types
    
    typedef V value_type;
    typedef A allocator_type;
    typedef V& reference;
    typedef V const& const_reference;
    typedef V* pointer;
    typedef V const* const_pointer;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;
    typedef I iterator;
    typedef J const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

    // structors
  
    explicit vector(A const& = A());
    explicit vector(size_type, V const& = V(), A const& = A());
  
    template <typename K>
    vector(K, K, A const& = A());
  
    vector(vector<V, A, R, I, J> const&);
    ~vector();
    vector<V, A, R, I, J>& operator=(vector<V, A, R, I, J> const&);

    void assign(size_type, V const&);

    template <typename K>
    void assign(K, K);

    // iterators

    iterator begin();
    const_iterator begin() const;
    iterator end();
    const_iterator end() const;
    reverse_iterator rbegin();
    const_reverse_iterator rbegin() const;
    reverse_iterator rend();
    const_reverse_iterator rend() const;

    // accessors

    A get_allocator() const;
    size_type size() const;
    size_type max_size() const;
    size_type capacity() const;
    bool empty() const;
    V const& operator[](size_type) const;
    V const& at(size_type) const;
    V const& front() const;
    V const& back() const;

    // modifiers

    V& operator[](size_type);
    V& at(size_type);
    V& front();
    V& back();
    void resize(size_type, V = V());
    void reserve(size_type);
    void push_back(V const&);
    void pop_back();
    iterator insert(iterator, V const&);
    void insert(iterator, size_type, V const&);

    template <typename K>
    void insert(iterator, K, K);

    iterator erase(iterator);
    iterator erase(iterator, iterator);
    void clear();
    void swap(vector<V, A, R, I, J>&);
  
  protected:

    template <typename K>
    void assign_dispatch(K, K, cphstl::int2type<0>);

    template <typename K>
    void assign_dispatch(K, K, cphstl::int2type<1>);

    void assign_fill(size_type, V const&);

    template <typename K>
    void assign_range(K, K);

    template <typename K>
    void insert_dispatch(iterator, K, K, cphstl::int2type<0>);

    template <typename K>
    void insert_dispatch(iterator, K, K, cphstl::int2type<1>);

    void insert_fill(iterator, size_type, V const&);

    template <typename K>
    void insert_range(iterator, K, K);

    typedef R realization_type;
    realization_type kernel; 

  };

  template <typename V, typename A, typename R, typename I, typename J>
  void swap(vector<V, A, R, I, J>&, vector<V, A, R, I, J>&);
  
  template <typename V, typename A, typename R, typename I, typename J>
  bool operator==(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);
  
  template <typename V, typename A, typename R, typename I, typename J>
  bool operator<(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);

  template <typename V, typename A, typename R, typename I, typename J>
  bool operator!=(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);
  
  template <typename V, typename A, typename R, typename I, typename J>
  bool operator>(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);

  template <typename V, typename A, typename R, typename I, typename J>
  bool operator>=(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);

  template <typename V, typename A, typename R, typename I, typename J>
  bool operator<=(vector<V, A, R, I, J> const&, vector<V, A, R, I, J> const&);

}

#include "vector.c++" // implements cphstl::vector

#endif
