#ifndef __CPHSTL_ALLOCATOR_PROXY__
#define __CPHSTL_ALLOCATOR_PROXY__

/*
  A generic allocator proxy class. It is actually more a real 
  encapsulator class as described in [1]. It keeps an instance 
  of the allocator, given by the parameterized constructor, 
  as a pointer. The allocator instance is allocated. The purpose of
  this class is to perform a safe swap operation for all containers. 
  Swap must not throw an exception, so when swapping this proxy class 
  we swap pointers which cannot fail. 
  All STL containers keeps an allocator which makes this class generic.

  [1] Pascoe, Geoffrey A. Encapsulators: a new software paradigm in 
      Smalltalk-80. SIGPLAN Notices 21(11), 1986, 341-346.

  Author: Bo Simonsen © 2009
*/

#include <algorithm>

namespace cphstl {
  template <typename A>
  class allocator_proxy {
  public:
    typedef typename A::size_type size_type;
    typedef typename A::difference_type difference_type;
    typedef typename A::pointer pointer;
    typedef typename A::const_pointer const_pointer;
    typedef typename A::reference reference;
    typedef typename A::const_reference const_reference;
    typedef typename A::value_type value_type;

    template <class U> 
    struct rebind { typedef allocator_proxy<typename A::template rebind<U>::other > other; };

    allocator_proxy() {
      (*this).a = new A();
    }
    allocator_proxy(A const& a) {
      (*this).a = new A(a);
    }
    ~allocator_proxy() {
      delete (*this).a;
    }
    allocator_proxy(allocator_proxy const& ap) {
      (*this).a = new A(*ap.a);
    }
    allocator_proxy(allocator_proxy& ap) {
      (*this).a = new A(*ap.a);
    }
    allocator_proxy operator=(allocator_proxy const& ap) {
      delete (*this).a;
      (*this).a = new A(*ap.a);
      return (*this);
    }
    allocator_proxy operator=(A const& a) {
      delete (*this).a;
      (*this).a = new A(a);
      return (*this);
    }
    pointer address(reference ref) const {
      return (*a).address(ref);
    }
    const_pointer address(const_reference const_ref) const {
      return (*a).address(const_ref);
    }
    size_type max_size() const throw() {
      return (*a).max_size();
    }
    pointer allocate(size_type s, const_pointer p = 0) {
      return (*a).allocate(s);
    }
    void construct(pointer p,value_type const& v) {
      (*a).construct(p,v);
    }
    void destroy(pointer p){
      (*a).construct(p);
    }
    void deallocate(pointer p,size_type s) {
      (*a).deallocate(p,s);
    }
    A& subject() const {
      return *a;
    }
  private:
    A* a;

  };

}

#endif
