#ifndef __CPHSTL_DOUBLY_INDIRECT_ENCAPSULATOR__
#define __CPHSTL_DOUBLY_INDIRECT_ENCAPSULATOR__

/*
  Defines a doubly indirect encapsulator.
  Segments store pointers to objects
  of the class doubly_indirect_encapsulator. Each
  object of doubly_indirect_encapsulator store a
  pointer to an object of the class value_encapsulator
  which stores the value.

  Authors: Bo Simonsen, May 2008, September 2008
  Bue Krogh Vedel-Larsen and Mikkel Jønsson Thomsen, March 2009
*/

#include <cassert>
#include <memory.h>

namespace cphstl {

  template <typename V>
  class value_encapsulator {

  public:
    typedef V value_type;

    value_encapsulator(value_type const& value)
      : value(value) {
    }

    value_type& content() {
      return value;
    }
    
    value_type const& content() const {
      return value;
    }

  private:
    value_type value;
  };

  template <typename V, typename A>
  class doubly_indirect_encapsulator;

  template <typename V, typename A>
  void slot_swap(doubly_indirect_encapsulator<V, A>*&, doubly_indirect_encapsulator<V, A>*&);

  template <typename V, typename A, bool is_const, typename E>
  class reference_proxy;

  template <typename V, typename A>
  class doubly_indirect_encapsulator {
  private:
    typedef doubly_indirect_encapsulator<V, A> this_type;
    typedef typename A::template 
      rebind<this_type>::other input_allocator_type;
  public:
    typedef V value_type;
    typedef A allocator_type;
    typedef std::size_t size_type;
    typedef this_type** segment_type;
    typedef this_type* slot_type;
    typedef value_encapsulator<V> value_encapsulator_type;
    typedef typename A::template 
      rebind<value_encapsulator_type>::other value_encapsulator_allocator_type;

    enum {size = sizeof(this_type*) + 
                 sizeof(value_encapsulator_type*) + 
                 sizeof(size_type) + 
                 sizeof(value_type)};

    friend void slot_swap<>(doubly_indirect_encapsulator<V, A>*&, doubly_indirect_encapsulator<V, A>*&);
    friend class reference_proxy<V, A, false, this_type>;
    friend class reference_proxy<V, A, true, this_type >;

  public:
    doubly_indirect_encapsulator(value_type const& _v, input_allocator_type& _a, size_type _index) {
      if ((*this).ap == 0) {
        (*this).a = _a;
        (*this).ap = &(*this).a;
      }

      (*this).v = (*this).a.allocate(1);
      new ((*this).v) value_encapsulator_type(_v);
      (*this).p = _index;
    }

    ~doubly_indirect_encapsulator() {
      (*(*this).v).~value_encapsulator_type();
      (*this).a.deallocate((*this).v, 1);
    }

    value_type& content() {
      return (*this).v->content();
    }

    value_type const& content() const {
      return (*this).v->content();
    }

    size_type position() const {
      return (*this).p;
    }
  private:
    static value_encapsulator_allocator_type a;
    static value_encapsulator_allocator_type* ap;
    size_type p;
    value_encapsulator_type* v;
  };

  template <typename V, typename A>
  typename doubly_indirect_encapsulator<V, A>::value_encapsulator_allocator_type* doubly_indirect_encapsulator<V, A>::ap = 0;

  template <typename V, typename A>
  typename doubly_indirect_encapsulator<V, A>::value_encapsulator_allocator_type doubly_indirect_encapsulator<V, A>::a;

}

#endif
