#ifndef __CPHSTL_DYNAMIC_ARRAY__
#define __CPHSTL_DYNAMIC_ARRAY__

/*
  The dynamic array kernel. 
  Expansion factor 2, Contraction threshold 1/4.
  If template argument fast is set, contraction threshold 0 will be used.

  Authors: Jyrki Katajainen, Bo Simonsen, May 2008
*/

#include <cstddef> // defines std::size_t
#include "direct-encapsulator.h++"
#include "type.h++"
#include "uninitialized-copy.i++"
#include "allocator-proxy.h++"

namespace cphstl {

  template <typename V, typename A, typename K>
  class vector_framework;

  template <
    typename V, 
    typename A,
    typename E = direct_encapsulator<V, A>,
    bool fast = false
  >
  class dynamic_array {

  public: 
    typedef std::size_t size_type;
    typedef E encapsulator_type;
    typedef typename E::segment_type segment_type;
    typedef typename E::slot_type slot_type;
    typedef typename A::template rebind<slot_type>::other segment_allocator_type;

    dynamic_array(A const& a = A()) 
      : segment_allocator(a), segment(0), current_capacity(0), current_size(0) {
    }

    ~dynamic_array() {
    }

    size_type max_size() const {
      typename A::template rebind<char>::other byte_allocator((*this).segment_allocator.subject());
      return byte_allocator.max_size() / (5 * sizeof(slot_type) + E::size);
    } 
 
    size_type size() const {
      return (*this).current_size;
    }

    void size(size_type s) {
      (*this).current_size = s;
    }

    size_type capacity() const {
      return (*this).current_capacity;
    }

    slot_type& access(size_type index) {
      return (*this).segment[index];
    }

    void grow(size_type delta) {
      size_type new_capacity = (*this).current_capacity == 0 ? 1 : (*this).current_capacity;
      segment_type new_segment = 0;

      do {
        new_capacity <<= 1;
      } while (new_capacity < (*this).current_size + delta);

      if (new_capacity >= (*this).max_size()) {
        new_capacity = (*this).current_size + delta;
      } 

      new_segment = (*this).segment_allocator.allocate(new_capacity);

      cphstl::uninitialized_copy((*this).segment, (*this).segment + (*this).current_size, new_segment);

      if ((*this).segment != 0) {
        (*this).segment_allocator.deallocate((*this).segment, (*this).current_capacity);
      }

      (*this).segment = new_segment;
      (*this).current_capacity = new_capacity;
    }

    void shrink() {
      if(fast) {
        return;
      }

      size_type new_capacity = (*this).current_capacity;

      while (new_capacity > (*this).current_size){
        new_capacity >>= 2;
      }
      new_capacity <<= 2;

      if ((*this).current_capacity == new_capacity) {
        return;
      }

      segment_type new_segment = 0;

      if(new_capacity != 0) {
        new_segment = segment_allocator.allocate(new_capacity); 
        cphstl::uninitialized_copy((*this).segment, (*this).segment + (*this).current_size, new_segment);
      }

      if ((*this).segment != 0) {
        segment_allocator.deallocate((*this).segment, (*this).current_capacity);
      }

      (*this).segment = new_segment;
      (*this).current_capacity = new_capacity;
    }

  private:
    allocator_proxy<segment_allocator_type> segment_allocator;
    segment_type segment;
    size_type current_capacity;
    size_type current_size;
  };
}

#endif
