#ifndef __CPHSTL_HASHED_ARRAY_TREE__
#define __CPHSTL_HASHED_ARRAY_TREE__

/*
  The hashed array tree kernel

  Authors: Jyrki Katajainen, Bo Simonsen © 2009
*/

#include <cmath>
#include <algorithm>

#include "allocator-proxy.h++"

namespace cphstl {
  template <typename V, typename A, typename K>
  class vector_framework;
  template <
    typename V, 
    typename A,
    typename E = direct_encapsulator<V, A>
  >
  class hashed_array_tree {

  public: 
    typedef std::size_t size_type;
    typedef E encapsulator_type;
    typedef typename E::segment_type segment_type;
    typedef typename E::slot_type slot_type;

  private:
    typedef segment_type* directory_type;
    typedef typename A::template rebind<slot_type>::other segment_allocator_type;
    typedef typename A::template rebind<segment_type>::other directory_allocator_type;

  public:
    hashed_array_tree(A const& a = A()) 
      : directory_allocator(a), segment_allocator(a), directory(0), 
        current_size(0), segment_size(0), number_of_segments(0), mask(0), 
        power(0) {
    }

    ~hashed_array_tree() {
    }

    size_type max_size() const {
      typename A::template rebind<char>::other byte_allocator((*this).segment_allocator.subject());
      float b = byte_allocator.max_size();
      float e = E::size;
      float p = sizeof(int*);
      float m = std::max(e, p);
      return size_type(floor(sqrt((- 2 * m - 4 + sqrt((m * 2 + 4) * (m * 2 + 4) + 8 * e * b)) / 4 * e)));
    }

    size_type size() const {
      return (*this).current_size;
    }

    void size(size_type s) {
      (*this).current_size = s;
    }

    size_type capacity() const {
      return (*this).segment_size * (*this).number_of_segments;
    }
    
    slot_type& access(size_type index) {
      return (*this).directory[index >> power][index & mask];
    }

    void grow(size_type delta) {
      if (((*this).current_size + delta > (*this).segment_size * (*this).segment_size)) {
        (*this).reorganize((*this).current_size + delta);
      }
      else {
        size_type new_number_of_segments = (*this).number_of_segments;
        try {
          while ((*this).get_number_of_segments((*this).current_size + delta, (*this).power, (*this).mask) > new_number_of_segments) {
            (*this).directory[new_number_of_segments] = segment_allocator.allocate((*this).segment_size);
            new_number_of_segments++;
          }
        } 
        catch (...) {
          while (new_number_of_segments > (*this).number_of_segments) {
            --new_number_of_segments;
            segment_allocator.deallocate((*this).directory[new_number_of_segments], (*this).segment_size);
          }
        }

        (*this).number_of_segments = new_number_of_segments;
      }
    }

    void shrink() {
      if (8 * (*this).size() < (*this).capacity()) {
        (*this).reorganize((*this).current_size);
      }
      else {
        while ((*this).get_number_of_segments((*this).current_size, (*this).power, (*this).mask) < (*this).number_of_segments) {
          --(*this).number_of_segments;
          segment_allocator.deallocate((*this).directory[(*this).number_of_segments], (*this).segment_size);
        }
        
        /*if((*this).number_of_segments == 0) {
          (*this).directory_allocator.deallocate((*this).directory, (*this).segment_size);
          (*this).segment_size = 0;
        }*/
      }
    }

  private:
    size_type get_number_of_segments(size_type required_capacity, size_type power, size_type mask) {
      return (required_capacity >> power) + ((required_capacity & mask) > 0 ? 1 : 0);
    } 

    void reorganize(size_type required_capacity) {
    
      size_type new_segment_size = 0;
      size_type new_power = 0;

      while (new_segment_size * new_segment_size < required_capacity) {
        new_power += 1;
        new_segment_size = 1 << new_power;
      }

      directory_type new_directory = 0;
      size_type new_mask = (1 << new_power) - 1;
      size_type new_number_of_segments = (*this).get_number_of_segments(required_capacity, new_power, new_mask);

      if(new_number_of_segments > 0) {
        new_directory = directory_allocator.allocate(new_segment_size);
      }

      size_type i = 0;

      try {
        while (i < new_number_of_segments) {
          new_directory[i] = segment_allocator.allocate(new_segment_size);
          ++i;
        }  
      } 
      catch (...) {
        while (i >= 0) {
          segment_allocator.deallocate(new_directory[i], new_segment_size);
          --i;
        }
        throw;
      }
      
      for (size_type i = 0; i < (*this).current_size; ++i) {
        new (&new_directory[i >> new_power][i & new_mask]) slot_type(directory[i >> power][i & mask]);
      }

      for (size_type i = 0; i < (*this).number_of_segments; ++i) {
        segment_allocator.deallocate(directory[i], (*this).segment_size);
      }

      directory_allocator.deallocate(directory, (*this).segment_size);

      (*this).segment_size = new_segment_size;
      (*this).power = new_power;
      (*this).mask = new_mask;
      (*this).number_of_segments = new_number_of_segments;
      (*this).directory = new_directory;
    }

    allocator_proxy<directory_allocator_type> directory_allocator;
    allocator_proxy<segment_allocator_type> segment_allocator;
    directory_type directory;
    size_type current_size;
    size_type segment_size;
    size_type number_of_segments;
    size_type mask;
    size_type power;
  };
}

#endif
