#ifndef __CPHSTL_LEVELWISE_ALLOCATED_PILE__
#define __CPHSTL_LEVELWISE_ALLOCATED_PILE__

/*
  The levelwise-allocated pile kernel. Notice, that the directory 
  is of a fixed size.  

  Authors: Filip Bruman, Jyrki Katajainen, Mads D. Kristensen, Bjarke
  Buur Mortensen, Daniel P. Larsen, Bo Simonsen, Christian Wolfgang ©
  2001, 2004, 2008, 2009
*/

#include <cmath> // defines ilogb

namespace cphstl {
  
  template <
    typename V, 
    typename A,
    typename E = direct_encapsulator<V, A>
  >
  class levelwise_allocated_pile {

  public: 
    typedef std::size_t size_type;
    typedef E encapsulator_type;
    typedef typename E::segment_type segment_type;
    typedef typename E::slot_type slot_type;
    typedef segment_type* directory_type;
    typedef typename A::template rebind<slot_type>::other segment_allocator_type;

    levelwise_allocated_pile(A const& a = A()) : segment_allocator(a), current_size(0), current_capacity(0), number_of_segments(0), last_segment_size(0) { 
    }

    ~levelwise_allocated_pile() {
    }

    size_type max_size() const {
      typename A::template rebind<char>::other byte_allocator((*this).segment_allocator.subject());
      size_type max_encapsulators = byte_allocator.max_size() / E::size;

      if (((*this).current_size + max_encapsulators) < max_encapsulators) {
        return size_type(-1);
      }

      return (*this).current_size + max_encapsulators;      
    }

    size_type size() const {
      return (*this).current_size;
    }

    void size(size_type s) {
      (*this).current_size = s;
    }

    size_type capacity() const {
      return (*this).current_capacity;
    }
    
    slot_type& access(size_type i) {
      size_type pow = ilogb(i+1);
      size_type index = i - ((1 << pow) - 1);
      return (*this).directory[pow][index];
    }

    void grow(size_type delta) {
      size_type new_capacity = (*this).current_capacity;
      size_type new_number_of_segments = (*this).number_of_segments;

      try {
        while ((*this).current_size + delta > new_capacity) {
          if ( new_capacity > (*this).max_size() ) {
            (*this).directory[new_number_of_segments] = segment_allocator.allocate((*this).current_size + delta);
            (*this).last_segment_size = (*this).current_size + delta;
            ++new_number_of_segments;	       
            new_capacity += delta;
            break;
          } else {
            new_capacity += 1 << new_number_of_segments;
            (*this).directory[new_number_of_segments] = segment_allocator.allocate(1 << new_number_of_segments);
            ++new_number_of_segments;	
          }
        }
      } 
      catch (...) {
        while (new_number_of_segments > (*this).number_of_segments) {
          --new_number_of_segments;
          segment_allocator.deallocate((*this).directory[new_number_of_segments], 1 << (new_number_of_segments));
        }
      }

      (*this).current_capacity = new_capacity;
      (*this).number_of_segments = new_number_of_segments;
       
    }

    void shrink() {
      while ( ((((*this).current_capacity - (1 << ((*this).number_of_segments - 1))) > (*this).current_size + 8) ||
              (*this).current_size == 0) && (*this).number_of_segments > 0) {
        if((*this).last_segment_size != 0) {
          --(*this).number_of_segments;
          (*this).segment_allocator.deallocate((*this).directory[(*this).number_of_segments], (*this).last_segment_size);
          (*this).current_capacity -= (*this).last_segment_size;   
          (*this).last_segment_size = 0;
        }
        else {
          --(*this).number_of_segments;
          (*this).segment_allocator.deallocate((*this).directory[(*this).number_of_segments], 1 << ((*this).number_of_segments));
          (*this).current_capacity -= ( 1 << ((*this).number_of_segments) );
        }
      }
    }

  private:
    segment_type directory[32];
    allocator_proxy<segment_allocator_type> segment_allocator;
    size_type current_size;
    size_type current_capacity;
    size_type number_of_segments;
    size_type last_segment_size;
  };
}

#endif
