#ifndef __CPHSTL_REFERENCE_PROXY__
#define __CPHSTL_REFERENCE_PROXY__

/*
  Defines the reference proxy. The framework will, instead
  of returning references, return an object of the reference
  proxy. 

  Author: Bo Simonsen © 2009
*/

#include "type.h++" // defines cphstl::if_then_else

namespace cphstl {

  template <typename V, typename A>
  class doubly_indirect_encapsulator;

  template <typename V, typename A>
  class indirect_encapsulator;

  template <typename V, typename A, bool is_const, typename E>
  class reference_proxy {

  public:
    typedef typename E::value_type value_type;
    typedef typename cphstl::if_then_else<is_const, E const&, E&>::type encapsulator_reference;    
    typedef typename cphstl::if_then_else<is_const, E const*, E*>::type encapsulator_pointer;    
    typedef typename cphstl::if_then_else<is_const, value_type const&, value_type&>::type reference_type;

    reference_proxy(value_type& _v) : e(0), v(_v) {
    }

    reference_proxy(value_type const& _v) : e(0), v(_v) {
    }    

    reference_proxy(encapsulator_reference _e) : e(&_e), v() {
    }

    reference_proxy(reference_proxy const& r) : e(r.e), v() {
    }

    value_type const& operator=(value_type const& _v) {
      if (e == 0) {
        (*this).v = _v;
        return (*this).v;
      }
      (*this).e->content() = _v;
      return (*this).e->content();
    }

    operator reference_type() {
      if (e == 0) {
        return v;
      }
      return (*this).e->content();
    }

  private:
    encapsulator_pointer e;
    V v;
  };

  template <typename V, typename A, bool is_const>
  class reference_proxy<V, A, is_const, indirect_encapsulator<V, A> > {

  public:
    typedef indirect_encapsulator<V, A> entry_type;
    typedef typename entry_type::value_type value_type;
    typedef typename cphstl::if_then_else<is_const, entry_type const*, entry_type*>::type entry_ptr;        
    typedef typename cphstl::if_then_else<is_const, value_type const&, value_type&>::type reference_type;
    
    reference_proxy(value_type& _v) : e(0), v(_v) {
    }

    reference_proxy(value_type const& _v) : e(0), v(_v) {
    }    
    
    reference_proxy(entry_ptr _e) : e(_e) {
    }

    reference_proxy(reference_proxy const& r) : e(r.e), v(r.v) {
    }

    value_type const& operator=(value_type const& _v) {
      if (e == 0) {
        (*this).v = _v;
        return (*this).v;
      }
      (*this).e->content() = _v;
      return (*this).e->content();
    }

    operator reference_type() {
      if ((*this).e == 0) {
        return v;
      }
      return (*this).e->content();
    }

  private:
    entry_ptr e;
    V v;

  };

  template <typename V, typename A, bool is_const>
  class reference_proxy<V, A, is_const, doubly_indirect_encapsulator<V, A> > {

  public:
    typedef doubly_indirect_encapsulator<V, A> encapsulator_type;
    typedef typename encapsulator_type::value_type value_type;
    typedef typename encapsulator_type::value_encapsulator_type value_encapsulator_type;
    typedef typename A::template rebind<value_encapsulator_type>::other allocator_type;
    typedef typename cphstl::if_then_else<is_const, encapsulator_type const*, encapsulator_type*>::type encapsulator_pointer;        
    typedef typename cphstl::if_then_else<is_const, value_type const&, value_type&>::type reference_type;    

    reference_proxy(encapsulator_pointer _e) : e(_e) {
    }

    reference_proxy(value_type& _v) : v(_v), e(0) {
    }

    reference_proxy(value_type const& _v) : v(_v), e(0) {
    }

    reference_proxy(reference_proxy const& r) : e(r.e) {
    }

    operator reference_type() {
      if ((*this).e == 0) {
        return (*this).v;
      }
      return (*this).e->content();
    }

    value_type const& operator=(value_type const& _v) {
      value_encapsulator_type* new_v;
      value_encapsulator_type* old_v;

      try {
        new_v = (*this).allocator.allocate(1);
        new (new_v) value_encapsulator_type(_v);

        old_v = (*(*this).e).v;
        (*(*this).e).v = new_v;

        (*old_v).~value_encapsulator_type();
        (*this).allocator.deallocate(old_v, 1);

      } 
      catch (...) {
        (*this).allocator.deallocate(new_v, 1);
        throw;
      }

      return _v;
    }

    reference_proxy& operator=(reference_proxy const& r) {
      (*this).e = r.e;
      return (*this);
    }

  private:
    allocator_type allocator;
    encapsulator_pointer e;
    value_type v;
  };

}
#endif
