#ifndef __CPHSTL_VECTOR_FRAMEWORK__
#define __CPHSTL_VECTOR_FRAMEWORK__

/*
  A skeleton of a vector implementation. Functionality is put into
  kernels and encapsulators. 

  Authors: Tina A. G. Andersen, Ulrik Schou Jørgensen, Jyrki
  Katajainen, Bo Simonsen, Mikkel Jønsson Thomsen, Claus Ullerlund,
  Bue Krogh Vedel-Larsen © 2008, 2009
*/

#include <cassert>
#include <cstddef> // std::size_t and std::ptrdiff_t
#include <iostream>
#include <memory> // std::allocator
#include <stdexcept> // std::out_of_range
#include <utility> // std::pair

#include "allocator-proxy.h++"
#include "dynamic-array.h++"
#include "encapsulator-factory.h++"
#include "general-surrogate.h++"
#include "reference-proxy.h++"
#include "slot-swap.i++"

namespace cphstl {

  template <
    typename V, 
    typename A = std::allocator<V>,
    typename K = dynamic_array<V, A>
  >
  class vector_framework {

  public:

    // types

    typedef V value_type;
    typedef A allocator_type;
    typedef K kernel_type;
    typedef std::size_t size_type;
    typedef general_surrogate<K> surrogate_type;
    typedef typename kernel_type::encapsulator_type encapsulator_type;
    typedef std::pair<size_type, surrogate_type*> concrete_iterator;
    typedef std::pair<size_type, surrogate_type*> concrete_const_iterator;
    typedef reference_proxy<V, A, false, encapsulator_type> reference;
    typedef reference_proxy<V, A, true, encapsulator_type> const_reference;
    typedef encapsulator_factory<V, A, encapsulator_type> factory_type;

    // structors

    explicit vector_framework(A const& = A());
    virtual ~vector_framework();
    
    // iterators

    concrete_iterator end();
    concrete_const_iterator end() const;
    concrete_iterator begin();
    concrete_const_iterator begin() const;
    
    // accessors

    A get_allocator() const;
    size_type size() const;
    size_type max_size() const;
    size_type capacity() const;
    const_reference operator[](size_type) const;
 
    // modifiers

    reference operator[](size_type);
    void reserve(size_type);
    concrete_iterator insert(concrete_iterator, V const&);
    void insert(concrete_iterator, size_type, V const&);
    
    template <typename IT>
    void insert(concrete_iterator, IT, IT);
    
    concrete_iterator erase(concrete_iterator);
    concrete_iterator erase(concrete_iterator, concrete_iterator);
    void swap(vector_framework<V, A, K>&);

  protected:

    size_type erase_values(concrete_iterator pos, size_type s);
    void block_copy_backward(size_type, size_type, size_type);
    void block_copy_forward(size_type, size_type, size_type);

  private:   

    typedef typename A::template rebind<surrogate_type>::other surrogate_allocator_type;

    allocator_proxy<surrogate_allocator_type> surrogate_allocator;
    factory_type f;
    A a;
    K k;
    surrogate_type* s;
  };
}

#include "vector-framework.i++"
#endif
