/*
  A rank-based iterator is just a (pointer, index) pair where the
  pointer points to the data structure containing the cell referred to
  and the index is the rank of that cell in the sequence of cells
  storing the elements.

  The idea of combining iterators and const iterators into the same
  class is taken from [Matt Austern. Defining iterators and const
  iterators. C/C++ User's Journal 19,1 (2001), 74-79].

  Authors: Jyrki Katajainen, Bo Simonsen (C) 2008
*/

#ifndef __CPHSTL_RANK_ITERATOR__
#define __CPHSTL_RANK_ITERATOR__

#include <cstddef> // defines std::size_t and std::ptrdiff_t
#include <iterator> // defines std::random_access_iterator_tag
#include "type.h++" // defines cphstl::if_then_else
#include <utility> // defines std::pair

namespace cphstl {

  template <typename V, typename A, typename R, typename I, typename J>
  class vector;

  template <typename R, bool is_const = false, typename E = typename R::encapsulator_type>
  class rank_iterator {
  #ifdef IS_ENCAPSULATOR_IN_FAMILY
    IS_ENCAPSULATOR_IN_FAMILY(E, rank)
  #endif

  public:
    // types

    typedef std::random_access_iterator_tag iterator_category;
    typedef typename R::value_type value_type;
    typedef std::size_t size_type;
    typedef std::ptrdiff_t difference_type;
    typedef typename if_then_else<is_const, value_type const*, value_type*>::type pointer;
    typedef typename if_then_else<is_const, typename R::const_reference, typename R::reference>::type reference;

    typedef E entry;
    typedef typename R::surrogate_type surrogate_type;

  protected:

    // types

    typedef typename if_then_else<is_const, E const*, E*>::type node_pointer;

  public:

    // friends

    friend class rank_iterator<R, !is_const, E>;

    template <typename V, typename A, typename S, typename I, typename J>
    friend class cphstl::vector;

    // structors

    rank_iterator();
    rank_iterator(rank_iterator<R, false, E> const&);
    rank_iterator(rank_iterator<R, true, E> const&);
    rank_iterator& operator=(rank_iterator const&);
    ~rank_iterator();

    // operators

    reference operator*() const;
    pointer operator->() const;
    rank_iterator& operator++();
    rank_iterator operator++(int);
    rank_iterator& operator--();
    rank_iterator operator--(int);
    rank_iterator& operator+=(difference_type);
    rank_iterator& operator-=(difference_type);
    rank_iterator operator+(difference_type) const;
    rank_iterator operator-(difference_type) const;
    difference_type operator-(rank_iterator const&) const;

    template <bool both>
    bool operator==(rank_iterator<R, both, E> const&) const;

    template <bool both>
    bool operator!=(rank_iterator<R, both, E> const&) const;

    template <bool both>
    bool operator<(rank_iterator<R, both, E> const&) const;

    template <bool both>
    bool operator>(rank_iterator<R, both, E> const&) const;

    template <bool both>
    bool operator<=(rank_iterator<R, both, E> const&) const;

    template <bool both>
    bool operator>=(rank_iterator<R, both, E> const&) const;

  protected:
    // converters to be used by the container friends
    rank_iterator(std::pair<size_type, surrogate_type*> const&);
    operator std::pair<size_type, surrogate_type*>() const;

    void advance(difference_type const& n);

    surrogate_type* surrogate;
    size_type position;  
  };

  template<typename R, bool both, typename E>
  rank_iterator<R, both, E> 
  operator+(typename R::difference_type, rank_iterator<R, both, E> const&);

}

#include "rank-iterator.i++" // implements cphstl::rank_iterator

#endif 
